#  Copyright (C) 2011  Statoil ASA, Norway.
#
#  The file 'ecl_grid.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

"""
Module to load and query ECLIPSE GRID/EGRID files.

The ecl_grid module contains functionality to load and query an
ECLIPSE grid file; it is currently not possible to manipulate or let
alone create a grid with ecl_grid module. The functionality is
implemented in the EclGrid class. The ecl_grid module is a thin
wrapper around the ecl_grid.c implementation from the libecl library.
"""
import ctypes

import warnings
import numpy
import sys
import os.path
import math
import itertools
from cwrap import CFILE, BaseCClass, load, open as copen

from ecl import EclPrototype
from ecl.util.util import monkey_the_camel
from ecl.util.util import IntVector
from ecl import  EclDataType, EclUnitTypeEnum
from ecl.eclfile import EclKW, FortIO
from ecl.grid import Cell


class EclGrid(BaseCClass):
    """
    Class for loading and internalizing ECLIPSE GRID/EGRID files.
    """

    TYPE_NAME = "ecl_grid"
    _fread_alloc                  = EclPrototype("void* ecl_grid_load_case__(char*, bool)", bind = False)
    _grdecl_create                = EclPrototype("ecl_grid_obj ecl_grid_alloc_GRDECL_kw(int, int, int, ecl_kw, ecl_kw, ecl_kw, ecl_kw)", bind = False)
    _alloc_rectangular            = EclPrototype("ecl_grid_obj ecl_grid_alloc_rectangular(int, int, int, double, double, double, int*)", bind = False)
    _exists                       = EclPrototype("bool ecl_grid_exists(char*)", bind = False)

    _get_numbered_lgr             = EclPrototype("ecl_grid_ref ecl_grid_get_lgr_from_lgr_nr(ecl_grid, int)")
    _get_named_lgr                = EclPrototype("ecl_grid_ref ecl_grid_get_lgr(ecl_grid, char*)")
    _get_cell_lgr                 = EclPrototype("ecl_grid_ref ecl_grid_get_cell_lgr1(ecl_grid, int)")
    _num_coarse_groups            = EclPrototype("int  ecl_grid_get_num_coarse_groups(ecl_grid)")
    _in_coarse_group1             = EclPrototype("bool ecl_grid_cell_in_coarse_group1(ecl_grid, int)")
    _free                         = EclPrototype("void ecl_grid_free(ecl_grid)")
    _get_nx                       = EclPrototype("int ecl_grid_get_nx(ecl_grid)")
    _get_ny                       = EclPrototype("int ecl_grid_get_ny(ecl_grid)")
    _get_nz                       = EclPrototype("int ecl_grid_get_nz(ecl_grid)")
    _get_global_size              = EclPrototype("int ecl_grid_get_global_size(ecl_grid)")
    _get_active                   = EclPrototype("int ecl_grid_get_active_size(ecl_grid)")
    _get_active_fracture          = EclPrototype("int ecl_grid_get_nactive_fracture(ecl_grid)")
    _get_name                     = EclPrototype("char* ecl_grid_get_name(ecl_grid)")
    _ijk_valid                    = EclPrototype("bool ecl_grid_ijk_valid(ecl_grid, int, int, int)")
    _get_active_index3            = EclPrototype("int ecl_grid_get_active_index3(ecl_grid, int, int, int)")
    _get_global_index3            = EclPrototype("int ecl_grid_get_global_index3(ecl_grid, int, int, int)")
    _get_active_index1            = EclPrototype("int ecl_grid_get_active_index1(ecl_grid, int)")
    _get_active_fracture_index1   = EclPrototype("int ecl_grid_get_active_fracture_index1(ecl_grid, int)")
    _get_global_index1A           = EclPrototype("int ecl_grid_get_global_index1A(ecl_grid, int)")
    _get_global_index1F           = EclPrototype("int ecl_grid_get_global_index1F(ecl_grid, int)")
    _get_ijk1                     = EclPrototype("void ecl_grid_get_ijk1(ecl_grid, int, int*, int*, int*)")
    _get_ijk1A                    = EclPrototype("void ecl_grid_get_ijk1A(ecl_grid, int, int*, int*, int*)")
    _get_xyz3                     = EclPrototype("void ecl_grid_get_xyz3(ecl_grid, int, int, int, double*, double*, double*)")
    _get_xyz1                     = EclPrototype("void ecl_grid_get_xyz1(ecl_grid, int, double*, double*, double*)")
    _get_cell_corner_xyz1         = EclPrototype("void ecl_grid_get_cell_corner_xyz1(ecl_grid, int, int, double*, double*, double*)")
    _get_corner_xyz               = EclPrototype("void ecl_grid_get_corner_xyz(ecl_grid, int, int, int, double*, double*, double*)")
    _get_xyz1A                    = EclPrototype("void ecl_grid_get_xyz1A(ecl_grid, int, double*, double*, double*)")
    _get_ij_xy                    = EclPrototype("bool ecl_grid_get_ij_from_xy(ecl_grid, double, double, int, int*, int*)")
    _get_ijk_xyz                  = EclPrototype("int  ecl_grid_get_global_index_from_xyz(ecl_grid, double, double, double, int)")
    _cell_contains                = EclPrototype("bool ecl_grid_cell_contains_xyz1(ecl_grid, int, double, double, double)")
    _cell_regular                 = EclPrototype("bool ecl_grid_cell_regular1(ecl_grid, int)")
    _num_lgr                      = EclPrototype("int  ecl_grid_get_num_lgr(ecl_grid)")
    _has_numbered_lgr             = EclPrototype("bool ecl_grid_has_lgr_nr(ecl_grid, int)")
    _has_named_lgr                = EclPrototype("bool ecl_grid_has_lgr(ecl_grid, char*)")
    _grid_value                   = EclPrototype("double ecl_grid_get_property(ecl_grid, ecl_kw, int, int, int)")
    _get_cell_volume              = EclPrototype("double ecl_grid_get_cell_volume1(ecl_grid, int)")
    _get_cell_thickness           = EclPrototype("double ecl_grid_get_cell_thickness1(ecl_grid, int)")
    _get_cell_dx                  = EclPrototype("double ecl_grid_get_cell_dx1(ecl_grid, int)")
    _get_cell_dy                  = EclPrototype("double ecl_grid_get_cell_dy1(ecl_grid, int)")
    _get_depth                    = EclPrototype("double ecl_grid_get_cdepth1(ecl_grid, int)")
    _fwrite_grdecl                = EclPrototype("void   ecl_grid_grdecl_fprintf_kw(ecl_grid, ecl_kw, char*, FILE, double)")
    _load_column                  = EclPrototype("void   ecl_grid_get_column_property(ecl_grid, ecl_kw, int, int, double_vector)")
    _get_top                      = EclPrototype("double ecl_grid_get_top2(ecl_grid, int, int)")
    _get_top1A                    = EclPrototype("double ecl_grid_get_top1A(ecl_grid, int)")
    _get_bottom                   = EclPrototype("double ecl_grid_get_bottom2(ecl_grid, int, int)")
    _locate_depth                 = EclPrototype("int    ecl_grid_locate_depth(ecl_grid, double, int, int)")
    _invalid_cell                 = EclPrototype("bool   ecl_grid_cell_invalid1(ecl_grid, int)")
    _valid_cell                   = EclPrototype("bool   ecl_grid_cell_valid1(ecl_grid, int)")
    _get_distance                 = EclPrototype("void   ecl_grid_get_distance(ecl_grid, int, int, double*, double*, double*)")
    _fprintf_grdecl2              = EclPrototype("void   ecl_grid_fprintf_grdecl2(ecl_grid, FILE, ecl_unit_enum) ")
    _fwrite_GRID2                 = EclPrototype("void   ecl_grid_fwrite_GRID2(ecl_grid, char*, ecl_unit_enum)")
    _fwrite_EGRID2                = EclPrototype("void   ecl_grid_fwrite_EGRID2(ecl_grid, char*, ecl_unit_enum)")
    _equal                        = EclPrototype("bool   ecl_grid_compare(ecl_grid, ecl_grid, bool, bool)")
    _dual_grid                    = EclPrototype("bool   ecl_grid_dual_grid(ecl_grid)")
    _init_actnum                  = EclPrototype("void   ecl_grid_init_actnum_data(ecl_grid, int*)")
    _compressed_kw_copy           = EclPrototype("void   ecl_grid_compressed_kw_copy(ecl_grid, ecl_kw, ecl_kw)")
    _global_kw_copy               = EclPrototype("void   ecl_grid_global_kw_copy(ecl_grid, ecl_kw, ecl_kw)")
    _create_volume_keyword        = EclPrototype("ecl_kw_obj ecl_grid_alloc_volume_kw(ecl_grid, bool)")
    _use_mapaxes                  = EclPrototype("bool ecl_grid_use_mapaxes(ecl_grid)")
    _export_coord                 = EclPrototype("ecl_kw_obj ecl_grid_alloc_coord_kw(ecl_grid)")
    _export_zcorn                 = EclPrototype("ecl_kw_obj ecl_grid_alloc_zcorn_kw(ecl_grid)")
    _export_actnum                = EclPrototype("ecl_kw_obj ecl_grid_alloc_actnum_kw(ecl_grid)")
    _export_mapaxes               = EclPrototype("ecl_kw_obj ecl_grid_alloc_mapaxes_kw(ecl_grid)")
    _get_unit_system              = EclPrototype("ecl_unit_enum ecl_grid_get_unit_system(ecl_grid)")



    @classmethod
    def load_from_grdecl(cls, filename):
        """Will create a new EclGrid instance from grdecl file.

        This function will scan the input file @filename and look for
        the keywords required to build a grid. The following keywords
        are required:

              SPECGRID   ZCORN   COORD

        In addition the function will look for and use the ACTNUM and
        MAPAXES keywords if they are found; if ACTNUM is not found all
        cells are assumed to be active.

        Slightly more exotic grid concepts like dual porosity, NNC
        mapping, LGR and coarsened cells will be completely ignored;
        if you need such concepts you must have an EGRID file and use
        the default EclGrid() constructor - that is also considerably
        faster.
        """

        if os.path.isfile(filename):
            with copen(filename) as f:
                specgrid = EclKW.read_grdecl(f, "SPECGRID", ecl_type=EclDataType.ECL_INT, strict=False)
                zcorn = EclKW.read_grdecl(f, "ZCORN")
                coord = EclKW.read_grdecl(f, "COORD")
                try:
                    actnum = EclKW.read_grdecl(f, "ACTNUM", ecl_type=EclDataType.ECL_INT)
                except ValueError:
                    actnum = None

                try:
                    mapaxes = EclKW.read_grdecl(f, "MAPAXES")
                except ValueError:
                    mapaxes = None

            return EclGrid.create(specgrid, zcorn, coord, actnum, mapaxes)
        else:
            raise IOError("No such file:%s" % filename)

    @classmethod
    def load_from_file(cls, filename):
        """
        Will inspect the @filename argument and create a new EclGrid instance.
        """
        if FortIO.isFortranFile(filename):
            return EclGrid(filename)
        else:
            return EclGrid.loadFromGrdecl(filename)


    @classmethod
    def create(cls, specgrid, zcorn, coord, actnum, mapaxes=None):

        """
        Create a new grid instance from existing keywords.

        This is a class method which can be used to create an EclGrid
        instance based on the EclKW instances @specgrid, @zcorn,
        @coord and @actnum. An ECLIPSE EGRID file contains the
        SPECGRID, ZCORN, COORD and ACTNUM keywords, so a somewhat
        involved way to create a EclGrid instance could be:

          file = ecl.EclFile("ECLIPSE.EGRID")
          specgrid_kw = file.iget_named_kw("SPECGRID", 0)
          zcorn_kw = file.iget_named_kw("ZCORN", 0)
          coord_kw = file.iget_named_kw("COORD", 0)
          actnum_kw = file.iget_named_kw("ACTNUM", 0)

          grid = EclGrid.create(specgrid_kw, zcorn_kw, coord_kw, actnum_kw)

        If you are so inclined ...
        """
        return cls._grdecl_create(specgrid[0], specgrid[1], specgrid[2], zcorn, coord, actnum, mapaxes)

    @classmethod
    def create_rectangular(cls, dims, dV, actnum=None):
        """
        Will create a new rectangular grid. @dims = (nx,ny,nz)  @dVg = (dx,dy,dz)

        With the default value @actnum == None all cells will be active,
        """

        warnings.warn("EclGrid.createRectangular is deprecated. " +
                "Please used the similar method in EclGridGenerator!",
                DeprecationWarning)

        if actnum is None:
            ecl_grid = cls._alloc_rectangular(dims[0], dims[1], dims[2], dV[0], dV[1], dV[2], None)
        else:
            if not isinstance(actnum, IntVector):
                tmp = IntVector(initial_size=len(actnum))
                for (index, value) in enumerate(actnum):
                    tmp[index] = value
                actnum = tmp

            if not len(actnum) == dims[0] * dims[1] * dims[2]:
                raise ValueError("ACTNUM size mismatch: len(ACTNUM):%d  Expected:%d" % (len(actnum), dims[0] * dims[1] * dims[2]))
            ecl_grid = cls._alloc_rectangular(dims[0], dims[1], dims[2], dV[0], dV[1], dV[2], actnum.getDataPtr())

        # If we have not succeeded in creatin the grid we *assume* the
        # error is due to a failed malloc.
        if ecl_grid is None:
            raise MemoryError("Failed to allocated regualar grid")

        return ecl_grid

    def __init__(self, filename, apply_mapaxes=True):
        """
        Will create a grid structure from an EGRID or GRID file.
        """
        c_ptr = self._fread_alloc(filename, apply_mapaxes)
        if c_ptr:
            super(EclGrid, self).__init__(c_ptr)
        else:
            raise IOError("Loading grid from:%s failed" % filename)


    def free(self):
        self._free()

    def _nicename(self):
        """name is often full path to grid, if so, output basename, else name"""
        name = self.getName()
        if os.path.isfile(name):
            name = os.path.basename(name)
        return name

    def __repr__(self):
        """Returns, e.g.:
           EclGrid("NORNE_ATW2013.EGRID", 46x112x22, global_size=113344, active_size=44431) at 0x28c4a70
        """
        name = self._nicename()
        if name:
            name = '"%s", ' % name
        g_size = self.getGlobalSize()
        a_size = self.getNumActive()
        xyz_s  = '%dx%dx%d' % (self.getNX(),self.getNY(),self.getNZ())
        return self._create_repr('%s%s, global_size=%d, active_size=%d' % (name, xyz_s, g_size, a_size))

    def __len__(self):
        """
        len(grid) wil return the total number of cells.
        """
        return self._get_global_size()

    def equal(self, other, include_lgr=True, include_nnc=False, verbose=False):
        """
        Compare the current grid with the other grid.
        """
        if not isinstance(other, EclGrid):
            raise TypeError("The other argument must be an EclGrid instance")
        return self._equal(other, include_lgr, include_nnc, verbose)


    def dual_grid(self):
        """Is this grid dual porosity model?"""
        return self._dual_grid()

    def get_dims(self):
        """A tuple of four elements: (nx, ny, nz, nactive)."""
        return (self.getNX(),
                 self.getNY(),
                 self.getNZ(),
                 self.getNumActive())


    @property
    def nx(self):
        return self._get_nx()

    def get_nx(self):
        """ The number of elements in the x direction"""
        return self._get_nx()

    @property
    def ny(self):
        return self._get_ny()

    def get_ny(self):
        """ The number of elements in the y direction"""
        return self._get_ny()

    @property
    def nz(self):
        return self._get_nz()

    def get_nz(self):
        """ The number of elements in the z direction"""
        return self._get_nz()

    def get_global_size(self):
        """Returns the total number of cells in this grid"""
        return self._get_global_size()

    def get_num_active(self):
        """The number of active cells in the grid."""
        return self._get_active()


    def get_num_active_fracture(self):
        """The number of active cells in the grid."""
        return self._get_active_fracture()


    def get_bounding_box_2d(self, layer=0, lower_left=None, upper_right=None):
        if 0 <= layer <= self.getNZ():
            x = ctypes.c_double()
            y = ctypes.c_double()
            z = ctypes.c_double()

            if lower_left is None:
                i1 = 0
                j1 = 0
            else:
                i1,j1 = lower_left
                if not 0 < i1 < self.getNX():
                    raise ValueError("lower_left i coordinate invalid")

                if not 0 < j1 < self.getNY():
                    raise ValueError("lower_left j coordinate invalid")


            if upper_right is None:
                i2 = self.getNX()
                j2 = self.getNY()
            else:
                i2,j2 = upper_right

                if not 1 < i2 <= self.getNX():
                    raise ValueError("upper_right i coordinate invalid")

                if not 1 < j2 <= self.getNY():
                    raise ValueError("upper_right j coordinate invalid")

            if not i1 < i2:
                raise ValueError("Must have lower_left < upper_right")

            if not j1 < j2:
                raise ValueError("Must have lower_left < upper_right")



            self._get_corner_xyz(i1, j1, layer, ctypes.byref(x), ctypes.byref(y), ctypes.byref(z))
            p0 = (x.value, y.value)

            self._get_corner_xyz(i2, j1, layer, ctypes.byref(x), ctypes.byref(y), ctypes.byref(z))
            p1 = (x.value, y.value )

            self._get_corner_xyz( i2, j2, layer, ctypes.byref(x), ctypes.byref(y), ctypes.byref(z))
            p2 = (x.value, y.value )

            self._get_corner_xyz(i1, j2, layer, ctypes.byref(x), ctypes.byref(y), ctypes.byref(z))
            p3 = (x.value, y.value )

            return (p0,p1,p2,p3)
        else:
            raise ValueError("Invalid layer value:%d  Valid range: [0,%d]" % (layer, self.getNZ()))


    def get_name(self):
        """
        Name of the current grid, returns a string.

        For the main grid this is the filename given to the
        constructor when loading the grid; for an LGR this is the name
        of the LGR. If the grid instance has been created with the
        create() classmethod this can be None.
        """
        n = self._get_name()
        return str(n) if n else ''

    def cell(self, global_index=None, active_index=None, i=None, j=None, k=None):
        if global_index is not None:
            return Cell(self, global_index)
        if active_index is not None:
            return Cell(self, self.global_index(active_index=active_index))
        if i is not None:
            return Cell(self, self.global_index(ijk=(i,j,k)))

    def __getitem__(self, global_index):
        if isinstance(global_index, tuple):
            i,j,k = global_index
            return self.cell(i=i, j=j, k=k)
        return self.cell(global_index=global_index)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def cells(self, active=False):
        """Iterator over all the (active) cells"""
        if not active:
            for c in self:
                yield c
        else:
            for i in range(self.get_num_active()):
                yield self.cell(active_index=i)

    def global_index(self, active_index=None, ijk=None):
        """
        Will convert either active_index or (i,j,k) to global index.
        """
        return self.__global_index(active_index=active_index, ijk=ijk)

    def __global_index(self, active_index=None, global_index=None, ijk=None):
        """
        Will convert @active_index or @ijk to global_index.

        This method will convert @active_index or @ijk to a global
        index. Exactly one of the arguments @active_index,
        @global_index or @ijk must be supplied.

        The method is used extensively internally in the EclGrid
        class; most methods which take coordinate input pass through
        this method to normalize the coordinate representation.
        """

        set_count = 0
        if not active_index is None:
            set_count += 1

        if not global_index is None:
            set_count += 1

        if ijk:
            set_count += 1

        if not set_count == 1:
            raise ValueError("Exactly one of the kewyord arguments active_index, global_index or ijk must be set")

        if not active_index is None:
            global_index = self._get_global_index1A( active_index)
        elif ijk:
            nx = self.getNX()
            ny = self.getNY()
            nz = self.getNZ()

            i,j,k = ijk

            if not 0 <= i < nx:
                raise IndexError("Invalid value i:%d  Range: [%d,%d)" % (i, 0, nx))

            if not 0 <= j < ny:
                raise IndexError("Invalid value j:%d  Range: [%d,%d)" % (j, 0, ny))

            if not 0 <= k < nz:
                raise IndexError("Invalid value k:%d  Range: [%d,%d)" % (k, 0, nz))

            global_index = self._get_global_index3(i,j,k)
        else:
            if not 0 <= global_index < self.getGlobalSize():
                raise IndexError("Invalid value global_index:%d  Range: [%d,%d)" % (global_index, 0, self.getGlobalSize()))
        return global_index


    def get_active_index(self, ijk=None, global_index=None):
        """
        Lookup active index based on ijk or global index.

        Will determine the active_index of a cell, based on either
        @ijk = (i,j,k) or @global_index. If the cell specified by the
        input arguments is not active the function will return -1.
        """
        gi = self.__global_index(global_index=global_index, ijk=ijk)
        return self._get_active_index1(gi)


    def get_active_fracture_index(self, ijk=None, global_index=None):
        """
        For dual porosity - get the active fracture index.
        """
        gi = self.__global_index(global_index=global_index, ijk=ijk)
        return self._get_active_fracture_index1(gi)


    def get_global_index1F(self, active_fracture_index):
        """
        Will return the global index corresponding to active fracture index.
        """
        return self._get_global_index1F(active_fracture_index)


    def cell_invalid(self, ijk=None, global_index=None, active_index=None):
        """
        Tries to check if a cell is invalid.

        Cells which are used to represent numerical aquifers are
        typically located in UTM position (0,0); these cells have
        completely whacked up shape and size, and should **NOT** be
        used in calculations involving real world coordinates. To
        protect against this a heuristic is used identify such cells
        and mark them as invalid. There might be other sources than
        numerical aquifers to this problem.
        """
        gi = self.__global_index(global_index=global_index, ijk=ijk, active_index=active_index)
        return self._invalid_cell(gi)


    def valid_cell_geometry(self, ijk=None, global_index=None, active_index=None):
        """Checks if the cell has valid geometry.

        There are at least two reasons why a cell might have invalid
        gemetry:

          1. In the case of GRID files it is not necessary to supply
             the geometry for all the cells; in that case this
             function will return false for cells which do not have
             valid coordinates.

          2. Cells which are used to represent numerical aquifers are
             typically located in UTM position (0,0); these cells have
             completely whacked up shape and size; these cells are
             identified by a heuristic - which might fail

        If the validCellGeometry() returns false for a particular
        cell functions which calculate cell volumes, real world
        coordinates and so on - should not be used.
        """
        gi = self.__global_index(global_index=global_index, ijk=ijk, active_index=active_index)
        return self._valid_cell(gi)



    def active(self, ijk=None, global_index=None):
        """
        Is the cell active?

        See documentation og get_xyz() for explanation of parameters
        @ijk and @global_index.
        """
        gi = self.__global_index(global_index=global_index, ijk=ijk)
        active_index = self._get_active_index1(gi)
        if active_index >= 0:
            return True
        else:
            return False


    def get_global_index(self, ijk=None, active_index=None):
        """
        Lookup global index based on ijk or active index.
        """
        gi = self.__global_index(active_index=active_index, ijk=ijk)
        return gi


    def get_ijk(self, active_index=None, global_index=None):
        """
        Lookup (i,j,k) for a cell, based on either active index or global index.

        The return value is a tuple with three elements (i,j,k).
        """
        i = ctypes.c_int()
        j = ctypes.c_int()
        k = ctypes.c_int()

        gi = self.__global_index(active_index=active_index, global_index=global_index)
        self._get_ijk1(gi, ctypes.byref(i), ctypes.byref(j), ctypes.byref(k))

        return (i.value, j.value, k.value)


    def get_xyz(self, active_index=None, global_index=None, ijk=None):
        """
        Find true position of cell center.

        Will return world position of the center of a cell in the
        grid. The return value is a tuple of three elements:
        (utm_x, utm_y, depth).

        The cells of a grid can be specified in three different ways:

           (i,j,k)      : As a tuple of i,j,k values.

           global_index : A number in the range [0,nx*ny*nz). The
                          global index is related to (i,j,k) as:

                            global_index = i + j*nx + k*nx*ny

           active_index : A number in the range [0,nactive).

        For many of the EclGrid methods a cell can be specified using
        any of these three methods. Observe that one and only method is
        allowed:

        OK:
            pos1 = grid.get_xyz(active_index=100)
            pos2 = grid.get_xyz(ijk=(10,20,7))

        Crash and burn:
            pos3 = grid.get_xyz(ijk=(10,20,7), global_index=10)
            pos4 = grid.get_xyz()

        All the indices in the EclGrid() class are zero offset, this
        is in contrast to ECLIPSE which has an offset 1 interface.
        """
        gi = self.__global_index(ijk=ijk, active_index=active_index, global_index=global_index)

        x = ctypes.c_double()
        y = ctypes.c_double()
        z = ctypes.c_double()
        self._get_xyz1(gi, ctypes.byref(x), ctypes.byref(y), ctypes.byref(z))
        return (x.value, y.value, z.value)


    def get_node_pos(self, i, j, k):
        """Will return the (x,y,z) for the node given by (i,j,k).

        Observe that this method does not consider cells, but the
        nodes in the grid. This means that the valid input range for
        i,j and k are are upper end inclusive. To get the four
        bounding points of the lower layer of the grid:

           p0 = grid.getNodePos(0, 0, 0)
           p1 = grid.getNodePos(grid.getNX(), 0, 0)
           p2 = grid.getNodePos(0, grid.getNY(), 0)
           p3 = grid.getNodePos(grid.getNX(), grid.getNY(), 0)

        """
        if not 0 <= i <= self.getNX():
            raise IndexError("Invalid I value:%d - valid range: [0,%d]" % (i, self.getNX()))

        if not 0 <= j <= self.getNY():
            raise IndexError("Invalid J value:%d - valid range: [0,%d]" % (j, self.getNY()))

        if not 0 <= k <= self.getNZ():
            raise IndexError("Invalid K value:%d - valid range: [0,%d]" % (k, self.getNZ()))

        x = ctypes.c_double()
        y = ctypes.c_double()
        z = ctypes.c_double()
        self._get_corner_xyz(i,j,k, ctypes.byref(x), ctypes.byref(y), ctypes.byref(z))
        return (x.value, y.value, z.value)


    def get_cell_corner(self, corner_nr, active_index=None, global_index=None, ijk=None):
        """
        Will look up xyz of corner nr @corner_nr


        lower layer:   upper layer

         2---3           6---7
         |   |           |   |
         0---1           4---5

        """
        gi = self.__global_index(ijk=ijk, active_index=active_index, global_index=global_index)
        x = ctypes.c_double()
        y = ctypes.c_double()
        z = ctypes.c_double()
        self._get_cell_corner_xyz1(gi, corner_nr, ctypes.byref(x), ctypes.byref(y), ctypes.byref(z))
        return (x.value, y.value, z.value)

    def get_node_xyz(self, i,j,k):
        """
        This function returns the position of Vertex (i,j,k).

        The coordinates are in the inclusive interval [0,nx] x [0,ny] x [0,nz].
        """
        nx = self.getNX()
        ny = self.getNY()
        nz = self.getNZ()

        corner = 0

        if i == nx:
            i -= 1
            corner += 1

        if j == ny:
            j -= 1
            corner += 2

        if k == nz:
            k -= 1
            corner += 4

        if self._ijk_valid(i, j, k):
            return self.getCellCorner(corner, global_index=i + j*nx + k*nx*ny)
        else:
            raise IndexError("Invalid coordinates: (%d,%d,%d) " % (i,j,k))



    def get_layer_xyz(self, xy_corner, layer):
        nx = self.getNX()

        (j, i) = divmod(xy_corner, nx + 1)
        k = layer
        return self.getNodeXYZ(i,j,k)



    def distance(self, global_index1, global_index2):
        dx = ctypes.c_double()
        dy = ctypes.c_double()
        dz = ctypes.c_double()
        self._get_distance(global_index1, global_index2, ctypes.byref(dx), ctypes.byref(dy), ctypes.byref(dz))
        return (dx.value, dy.value, dz.value)


    def depth(self, active_index=None, global_index=None, ijk=None):
        """
        Depth of the center of a cell.

        Returns the depth of the center of the cell given by
        @active_index, @global_index or @ijk. See method get_xyz() for
        documentation of @active_index, @global_index and @ijk.
        """
        gi = self.__global_index(ijk=ijk, active_index=active_index, global_index=global_index)
        return self._get_depth( gi)

    def top(self, i, j):
        """
        Top of the reservoir; in the column (@i, @j).
        Returns average depth of the four top corners.
        """
        return self._get_top(i, j)

    def top_active(self, i, j):
        """
        Top of the active part of the reservoir; in the column (@i, @j).
        Raises ValueError if (i,j) column is inactive.
        """
        for k in range(self.getNZ()):
            a_idx = self.get_active_index(ijk=(i,j,k))
            if a_idx >= 0:
                return self._get_top1A(a_idx)
        raise ValueError('No active cell in column (%d,%d)' % (i,j))

    def bottom(self, i, j):
        """
        Bottom of the reservoir; in the column (@i, @j).
        """
        return self._get_bottom( i, j)

    def locate_depth(self, depth, i, j):
        """
        Will locate the k value of cell containing specified depth.

        Will scan through the grid column specified by the input
        arguments @i and @j and search for a cell containing the depth
        given by input argument @depth. The return value is the k
        value of cell containing @depth.

        If @depth is above the top of the reservoir the function will
        return -1, and if @depth is below the bottom of the reservoir
        the function will return -nz.
        """
        return self._locate_depth( depth, i, j)


    def find_cell(self, x, y, z, start_ijk=None):
        """
        Lookup cell containg true position (x,y,z).

        Will locate the cell in the grid which contains the true
        position (@x,@y,@z), the return value is as a triplet
        (i,j,k). The underlying C implementation is not veeery
        efficient, and can potentially take quite long time. If you
        provide a good intial guess with the parameter @start_ijk (a
        tuple (i,j,k)) things can speed up quite substantially.

        If the location (@x,@y,@z) can not be found in the grid, the
        method will return None.
        """
        start_index = 0
        if start_ijk:
            start_index = self.__global_index(ijk=start_ijk)

        global_index = self._get_ijk_xyz(x, y, z, start_index)
        if global_index >= 0:
            i = ctypes.c_int()
            j = ctypes.c_int()
            k = ctypes.c_int()
            self._get_ijk1(global_index,
                            ctypes.byref(i), ctypes.byref(j), ctypes.byref(k))
            return (i.value, j.value, k.value)
        return None

    def cell_contains(self, x, y, z, active_index=None, global_index=None, ijk=None):
        """
        Will check if the cell contains point given by world
        coordinates (x,y,z).

        See method get_xyz() for documentation of @active_index,
        @global_index and @ijk.
        """
        gi = self.__global_index(ijk=ijk, active_index=active_index, global_index=global_index)
        return self._cell_contains(gi, x,y,z)


    def find_cell_xy(self, x, y, k):
        """Will find the i,j of cell with utm coordinates x,y.

        The @k input is the layer you are interested in, the allowed
        values for k are [0,nz]. If the coordinates (x,y) are found to
        be outside the grid a ValueError exception is raised.
        """
        if 0 <= k <= self.getNZ():
            i = ctypes.c_int()
            j = ctypes.c_int()
            ok = self._get_ij_xy(x,y,k, ctypes.byref(i), ctypes.byref(j))
            if ok:
                return (i.value, j.value)
            else:
                raise ValueError("Could not find the point:(%g,%g) in layer:%d" % (x,y,k))
        else:
            raise IndexError("Invalid layer value:%d" % k)

    def find_cell_corner_xy(self, x, y, k):
        """Will find the corner nr of corner closest to utm coordinates x,y.

        The @k input is the layer you are interested in, the allowed
        values for k are [0,nz]. If the coordinates (x,y) are found to
        be outside the grid a ValueError exception is raised.
        """
        i,j = self.findCellXY(x,y,k)
        if k == self.getNZ():
            k -= 1
            corner_shift = 4
        else:
            corner_shift = 0

        nx = self.getNX()
        x0,y0,z0 = self.getCellCorner(corner_shift, ijk=(i,j,k))
        d0 = math.sqrt((x0 - x)*(x0 - x) + (y0 - y)*(y0 - y))
        c0 = i + j*(nx + 1)

        x1,y1,z1 = self.getCellCorner(1 + corner_shift, ijk=(i,j,k))
        d1 = math.sqrt((x1 - x)*(x1 - x) + (y1 - y)*(y1 - y))
        c1 = i + 1 + j*(nx + 1)

        x2,y2,z2 = self.getCellCorner(2 + corner_shift, ijk=(i,j,k))
        d2 = math.sqrt((x2 - x)*(x2 - x) + (y2 - y)*(y2 - y))
        c2 = i + (j + 1)*(nx + 1)

        x3,y3,z3 = self.getCellCorner(3 + corner_shift, ijk=(i,j,k))
        d3 = math.sqrt((x3 - x)*(x3 - x) + (y3 - y)*(y3 - y))
        c3 = i + 1 + (j + 1)*(nx + 1)

        l = [(d0, c0), (d1,c1), (d2, c2), (d3,c3)]
        l.sort(key=lambda k: k[0])
        return l[0][1]



    def cell_regular(self, active_index=None, global_index=None, ijk=None):
        """
        The ECLIPSE grid models often contain various degenerate cells,
        which are twisted, have overlapping corners or what not. This
        function gives a moderate sanity check on a cell, essentially
        what the function does is to check if the cell contains it's
        own centerpoint - which is actually not as trivial as it
        sounds.
        """
        gi = self.__global_index(ijk=ijk, active_index=active_index, global_index=global_index)
        return self._cell_regular( gi)


    def cell_volume(self, active_index=None, global_index=None, ijk=None):
        """
        Calculate the volume of a cell.

        Will calculate the total volume of the cell. See method
        get_xyz() for documentation of @active_index, @global_index
        and @ijk.
        """
        gi = self.__global_index(ijk=ijk, active_index=active_index, global_index=global_index)
        return self._get_cell_volume(gi)


    def cell_dz(self, active_index=None, global_index=None, ijk=None):
        """
        The thickness of a cell.

        Will calculate the (average) thickness of the cell. See method
        get_xyz() for documentation of @active_index, @global_index
        and @ijk.
        """
        gi = self.__global_index(ijk=ijk, active_index=active_index, global_index=global_index)
        return self._get_cell_thickness( gi)


    def get_cell_dims(self, active_index=None, global_index=None, ijk=None):
        """Will return a tuple (dx,dy,dz) for cell dimension.

        The dx and dy values are best effor estimates of the cell size
        along the i and j directions respectively. The three values
        are guaranteed to satisfy:

              dx * dy * dz = dV

        See method get_xyz() for documentation of @active_index,
        @global_index and @ijk.

        """
        gi = self.__global_index(ijk=ijk, active_index=active_index, global_index=global_index)
        dx = self._get_cell_dx(gi)
        dy = self._get_cell_dy(gi)
        dz = self._get_cell_thickness( gi)
        return (dx,dy,dz)



    def get_num_lgr(self):

        """
        How many LGRs are attached to this main grid?

        How many LGRs are attached to this main grid; the grid
        instance doing the query must itself be a main grid.
        """
        return self._num_lgr()



    def has_lgr(self, lgr_name):
        """
        Query if the grid has an LGR with name @lgr_name.
        """
        if self._has_named_lgr(lgr_name):
            return True
        else:
            return False


    def get_lgr(self, lgr_key):
        """Get EclGrid instance with LGR content.

        Return an EclGrid instance based on the LGR @lgr, the input
        argument can either be the name of an LGR or the grid number
        of the LGR. The LGR grid instance is mostly like an ordinary
        grid instance; the only difference is that it can not be used
        for further queries about LGRs.

        If the grid does not contain an LGR with this name/nr
        exception KeyError will be raised.

        """
        lgr = None
        if isinstance(lgr_key, int):
            if self._has_numbered_lgr(lgr_key):
                lgr = self._get_numbered_lgr(lgr_key)
        else:
            if self._has_named_lgr(lgr_key):
                lgr = self._get_named_lgr(lgr_key)

        if lgr is None:
            raise KeyError("No such LGR: %s" % lgr_key)

        lgr.setParent(self)
        return lgr



    def get_cell_lgr(self, active_index=None, global_index=None, ijk=None):
        """
        Get EclGrid instance located in cell.

        Will query the current grid instance if the cell given by
        @active_index, @global_index or @ijk has been refined with an
        LGR. Will return None if the cell in question has not been
        refined, the return value can be used for further queries.

        See get_xyz() for documentation of the input parameters.
        """
        gi  = self.__global_index(ijk=ijk, active_index=active_index, global_index=global_index)
        lgr = self._get_cell_lgr(gi)
        if lgr:
            lgr.setParent(self)
            return lgr
        else:
            raise IndexError("No LGR defined for this cell")


    def grid_value(self, kw, i, j, k):
        """
        Will evalute @kw in location (@i,@j,@k).

        The ECLIPSE properties and solution vectors are stored in
        restart and init files as 1D vectors of length nx*nx*nz or
        nactive. The grid_value() method is a minor convenience
        function to convert the (@i,@j,@k) input values to an
        appropriate 1D index.

        Depending on the length of kw the input arguments are
        converted either to an active index or to a global index. If
        the length of kw does not fit with either the global size of
        the grid or the active size of the grid things will fail hard.
        """
        return self._grid_value(kw, i, j, k)


    def load_column(self, kw, i, j, column):
        """
        Load the values of @kw from the column specified by (@i,@j).

        The method will scan through all k values of the input field
        @kw for fixed values of i and j. The size of @kw must be
        either nactive or nx*ny*nz.

        The input argument @column should be a DoubleVector instance,
        observe that if size of @kw == nactive k values corresponding
        to inactive cells will not be modified in the @column
        instance; in that case it is important that @column is
        initialized with a suitable default value.
        """
        self._load_column( kw, i, j, column)


    def create_kw(self, array, kw_name, pack):
        """
        Creates an EclKW instance based on existing 3D numpy object.

        The method create3D() does the inverse operation; creating a
        3D numpy object from an EclKW instance. If the argument @pack
        is true the resulting keyword will have length 'nactive',
        otherwise the element will have length nx*ny*nz.
        """
        if array.ndim == 3:
            dims = array.shape
            if dims[0] == self.getNX() and dims[1] == self.getNY() and dims[2] == self.getNZ():
                dtype = array.dtype
                if dtype == numpy.int32:
                    type = EclDataType.ECL_INT
                elif dtype == numpy.float32:
                    type = EclDataType.ECL_FLOAT
                elif dtype == numpy.float64:
                    type = EclDataType.ECL_DOUBLE
                else:
                    sys.exit("Do not know how to create ecl_kw from type:%s" % dtype)

                if pack:
                    size = self.getNumActive()
                else:
                    size = self.getGlobalSize()

                if len(kw_name) > 8:
                    # Silently truncate to length 8 - ECLIPSE has it's challenges.
                    kw_name = kw_name[0:8]

                kw = EclKW(kw_name, size, type)
                active_index = 0
                global_index = 0
                for k in range(self.getNZ()):
                    for j in range(self.getNY()):
                        for i in range(self.getNX()):
                            if pack:
                                if self.active(global_index=global_index):
                                    kw[active_index] = array[i,j,k]
                                    active_index += 1
                            else:
                                if dtype == numpy.int32:
                                    kw[global_index] = int(array[i,j,k])
                                else:
                                    kw[global_index] = array[i,j,k]

                            global_index += 1
                return kw
        raise ValueError("Wrong size / dimension on array")


    def coarse_groups(self):
        """
        Will return the number of coarse groups in this grid.
        """
        return self._num_coarse_groups()


    def in_coarse_group(self, global_index=None, ijk=None, active_index=None):
        """
        Will return True or False if the cell is part of coarse group.
        """
        global_index = self.__global_index(active_index=active_index, ijk=ijk, global_index=global_index)
        return self._in_coarse_group1(global_index)


    def create_3d(self, ecl_kw, default = 0):
        """
        Creates a 3D numpy array object with the data from  @ecl_kw.

        Observe that 3D numpy object is a copy of the data in the
        EclKW instance, i.e. modification to the numpy object will not
        be reflected in the ECLIPSE keyword.

        The methods createKW() does the inverse operation; creating an
        EclKW instance from a 3D numpy object.

        Alternative: Creating the numpy array object is not very
        efficient; if you only need a limited number of elements from
        the ecl_kw instance it might be wiser to use the grid_value()
        method:

           value = grid.grid_value(ecl_kw, i, j, k)

        """
        if len(ecl_kw) == self.getNumActive() or len(ecl_kw) == self.getGlobalSize():
            array = numpy.ones([ self.getGlobalSize() ], dtype=ecl_kw.dtype) * default
            kwa = ecl_kw.array
            if len(ecl_kw) == self.getGlobalSize():
                for i in range(kwa.size):
                    array[i] = kwa[i]
            else:
                data_index = 0
                for global_index in range(self.getGlobalSize()):
                    if self.active(global_index=global_index):
                        array[global_index] = kwa[data_index]
                        data_index += 1

            array = array.reshape([self.getNX(), self.getNY(), self.getNZ()], order='F')
            return array
        else:
            err_msg_fmt = 'Keyword "%s" has invalid size %d; must be either nactive=%d or nx*ny*nz=%d'
            err_msg = err_msg_fmt % (ecl_kw, len(ecl_kw), self.getNumActive(),
                                     self.getGlobalSize())
            raise ValueError(err_msg)

    def save_grdecl(self, pyfile, output_unit=EclUnitTypeEnum.ECL_METRIC_UNITS):
        """
        Will write the the grid content as grdecl formatted keywords.

        Will only write the main grid.
        """
        cfile = CFILE(pyfile)
        self._fprintf_grdecl2(cfile, output_unit)

    def save_EGRID(self, filename, output_unit=None):
         if output_unit is None:
             output_unit = self.unit_system
         self._fwrite_EGRID2(filename, output_unit)

    def save_GRID(self, filename, output_unit=EclUnitTypeEnum.ECL_METRIC_UNITS):
        """
        Will save the current grid as a GRID file.
        """
        self._fwrite_GRID2( filename, output_unit)


    def write_grdecl(self, ecl_kw, pyfile, special_header=None, default_value=0):
        """
        Writes an EclKW instance as an ECLIPSE grdecl formatted file.

        The input argument @ecl_kw must be an EclKW instance of size
        nactive or nx*ny*nz. If the size is nactive the inactive cells
        will be filled with @default_value; hence the function will
        always write nx*ny*nz elements.

        The data in the @ecl_kw argument can be of type integer,
        float, double or bool. In the case of bool the default value
        must be specified as 1 (True) or 0 (False).

        The input argument @pyfile should be a valid python filehandle
        opened for writing; i.e.

           pyfile = open("PORO.GRDECL", "w")
           grid.write_grdecl(poro_kw , pyfile, default_value=0.0)
           grid.write_grdecl(permx_kw, pyfile, default_value=0.0)
           pyfile.close()

        """

        if len(ecl_kw) == self.getNumActive() or len(ecl_kw) == self.getGlobalSize():
            cfile = CFILE(pyfile)
            self._fwrite_grdecl(ecl_kw, special_header, cfile, default_value)
        else:
            raise ValueError("Keyword: %s has invalid size(%d), must be either nactive:%d  or nx*ny*nz:%d" % (ecl_kw.getName(), len(ecl_kw), self.getNumActive(), self.getGlobalSize()))


    def exportACTNUM(self):
        actnum = IntVector(initial_size=self.getGlobalSize())
        self._init_actnum(actnum.getDataPtr())
        return actnum


    def compressed_kw_copy(self, kw):
        if len(kw) == self.getNumActive():
            return kw.copy()
        elif len(kw) == self.getGlobalSize():
            kw_copy = EclKW(kw.getName(), self.getNumActive(), kw.data_type)
            self._compressed_kw_copy(kw_copy, kw)
            return kw_copy
        else:
            raise ValueError("The input keyword must have nx*n*nz or nactive elements. Size:%d invalid" % len(kw))

    def global_kw_copy(self, kw, default_value):
        if len(kw) == self.getGlobalSize():
            return kw.copy()
        elif len(kw) == self.getNumActive():
            kw_copy = EclKW(kw.getName(), self.getGlobalSize(), kw.data_type)
            kw_copy.assign(default_value)
            self._global_kw_copy(kw_copy, kw)
            return kw_copy
        else:
            raise ValueError("The input keyword must have nx*n*nz or nactive elements. Size:%d invalid" % len(kw))


    def export_ACTNUM_kw(self):
        actnum = EclKW("ACTNUM", self.getGlobalSize(), EclDataType.ECL_INT)
        self._init_actnum(actnum.getDataPtr())
        return actnum


    def create_volume_keyword(self, active_size=True):
        """Will create a EclKW initialized with cell volumes.

        The purpose of this method is to create a EclKW instance which
        is initialized with all the cell volumes, this can then be
        used to perform volume summation; i.e. to calculate the total
        oil volume:

           soil = 1 - sgas - swat
           cell_volume = grid.createVolumeKeyword()
           tmp = cell_volume * soil
           oip = tmp.sum()

        The oil in place calculation shown above could easily be
        implemented by iterating over the soil kw, however using the
        volume keyword has two advantages:

          1. The calculation of cell volumes is quite time consuming,
             by storing the results in a kw they can be reused.

          2. By using the compact form 'oip = cell_volume * soil' the
             inner loop iteration will go in C - which is faster.

        By default the kw will only have values for the active cells,
        but by setting the optional variable @active_size to False you
        will get volume values for all cells in the grid.
        """

        return self._create_volume_keyword(active_size)

    def export_coord(self):
        return self._export_coord()

    def export_zcorn(self):
        return self._export_zcorn()

    def export_actnum(self):
        return self._export_actnum()

    def export_mapaxes(self):
        if not self._use_mapaxes():
            return None

        return self._export_mapaxes()

    @property
    def unit_system(self):
        return self._get_unit_system()

monkey_the_camel(EclGrid, 'loadFromGrdecl', EclGrid.load_from_grdecl, classmethod)
monkey_the_camel(EclGrid, 'loadFromFile', EclGrid.load_from_file, classmethod)
monkey_the_camel(EclGrid, 'createRectangular', EclGrid.create_rectangular, classmethod)
monkey_the_camel(EclGrid, 'dualGrid', EclGrid.dual_grid)
monkey_the_camel(EclGrid, 'getDims', EclGrid.get_dims)
monkey_the_camel(EclGrid, 'getNX', EclGrid.get_nx)
monkey_the_camel(EclGrid, 'getNY', EclGrid.get_ny)
monkey_the_camel(EclGrid, 'getNZ', EclGrid.get_nz)
monkey_the_camel(EclGrid, 'getGlobalSize', EclGrid.get_global_size)
monkey_the_camel(EclGrid, 'getNumActive', EclGrid.get_num_active)
monkey_the_camel(EclGrid, 'getNumActiveFracture', EclGrid.get_num_active_fracture)
monkey_the_camel(EclGrid, 'getBoundingBox2D', EclGrid.get_bounding_box_2d)
monkey_the_camel(EclGrid, 'getName', EclGrid.get_name)
monkey_the_camel(EclGrid, 'validCellGeometry', EclGrid.valid_cell_geometry)
monkey_the_camel(EclGrid, 'getNodePos', EclGrid.get_node_pos)
monkey_the_camel(EclGrid, 'getCellCorner', EclGrid.get_cell_corner)
monkey_the_camel(EclGrid, 'getNodeXYZ', EclGrid.get_node_xyz)
monkey_the_camel(EclGrid, 'getLayerXYZ', EclGrid.get_layer_xyz)
monkey_the_camel(EclGrid, 'findCellXY', EclGrid.find_cell_xy)
monkey_the_camel(EclGrid, 'findCellCornerXY', EclGrid.find_cell_corner_xy)
monkey_the_camel(EclGrid, 'getCellDims', EclGrid.get_cell_dims)
monkey_the_camel(EclGrid, 'getNumLGR', EclGrid.get_num_lgr)
monkey_the_camel(EclGrid, 'createKW', EclGrid.create_kw)
monkey_the_camel(EclGrid, 'create3D', EclGrid.create_3d)
monkey_the_camel(EclGrid, 'compressedKWCopy', EclGrid.compressed_kw_copy)
monkey_the_camel(EclGrid, 'globalKWCopy', EclGrid.global_kw_copy)
monkey_the_camel(EclGrid, 'exportACTNUMKw', EclGrid.export_ACTNUM_kw)
monkey_the_camel(EclGrid, 'createVolumeKeyword', EclGrid.create_volume_keyword)



from cwrap import BaseCClass
from ecl import EclPrototype

class WellSegment(BaseCClass):
    TYPE_NAME = "well_segment"

    _active           = EclPrototype("bool well_segment_active(well_segment)")
    _main_stem        = EclPrototype("bool well_segment_main_stem(well_segment)")
    _nearest_wellhead = EclPrototype("bool well_segment_nearest_wellhead(well_segment)")
    _id               = EclPrototype("int well_segment_get_id(well_segment)")
    _link_count       = EclPrototype("int well_segment_get_link_count(well_segment)")
    _branch_id        = EclPrototype("int well_segment_get_branch_id(well_segment)")
    _outlet_id        = EclPrototype("int well_segment_get_outlet_id(well_segment)")
    _depth            = EclPrototype("double well_segment_get_depth(well_segment)")
    _length           = EclPrototype("double well_segment_get_length(well_segment)")
    _total_length     = EclPrototype("double well_segment_get_total_length(well_segment)")
    _diameter         = EclPrototype("double well_segment_get_diameter(well_segment)")

    def __init__(self):
        raise NotImplementedError("Class can not be instantiated directly")

    def free(self):
        pass

    def __repr__(self):
        return 'WellSegment(%s) at 0x%x' % (str(self), self._address())

    def __str__(self):
        return "{Segment ID:%d   BranchID:%d  Length:%g}" % (self.id() , self.branchId() , self.length())

    def id(self):
        """ @rtype: int """
        return self._id()

    def linkCount(self):
        """ @rtype: int """
        return self._link_count()

    def branchId(self):
        """ @rtype: int """
        return self._branch_id()

    def outletId(self):
        """ @rtype: int """
        return self._outlet_id()

    def isActive(self):
        """ @rtype: bool """
        return self._active()

    def isMainStem(self):
        """ @rtype: bool """
        return self._main_stem()

    def isNearestWellHead(self):
        """ @rtype: bool """
        return self._nearest_wellhead()

    def depth(self):
        """ @rtype: float """
        return self._depth()

    def __len__(self):
        return self.length()

    def length(self):
        """ @rtype: float """
        return self._length()

    def totalLength(self):
        """ @rtype: float """
        return self._total_length()

    def diameter(self):
        """ @rtype: float """
        return self._diameter()



#  Copyright (C) 2011  Statoil ASA, Norway.
#
#  The file 'ecl_kw.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
"""
Support for working with one keyword from ECLIPSE file.

ECLIPSE files in "restart format" are organized in keywords consisting
of a simple header and blocks of data. A keyword typically looks like:

  'SWAT    '  10000  'REAL'
  0.05  0.08  0.08  0.10
  0.11  0.11  0.10  0.09
  ....

I.e. it starts with of header consisting of a 8 characters name, a
length and a datatype, immediately followed by the actual
data.

Altough the term "restart format" is used to describe the format, this
particular format is not limited to restart files; it is (at least)
used in INIT, EGRID, GRID, Snnn, UNSMRY, SMSPEC, UNRST, Xnnnn and RFT
files. This module also has (some) support for working with GRDECL
'formatted' files.

The ecl_kw.py implementation wraps the ecl_kw.c implementation from
the libecl library.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import ctypes
import warnings
import numpy

from cwrap import CFILE, BaseCClass

from ecl import EclPrototype
from ecl.util.util import monkey_the_camel
from ecl import EclDataType
from ecl import EclTypeEnum, EclUtil

from .fortio import FortIO

def dump_type_deprecation_warning():
    warnings.warn("EclTypeEnum is deprecated. " +
        "You should instead provide an EclDataType",
        DeprecationWarning)

def constant_size_data_type(ecl_type):
    return (ecl_type in [
                        EclTypeEnum.ECL_CHAR_TYPE,
                        EclTypeEnum.ECL_FLOAT_TYPE,
                        EclTypeEnum.ECL_DOUBLE_TYPE,
                        EclTypeEnum.ECL_INT_TYPE,
                        EclTypeEnum.ECL_BOOL_TYPE,
                        EclTypeEnum.ECL_MESS_TYPE
                        ])

def warn_and_cast_data_type(data_type):
    if isinstance(data_type, EclDataType):
        return data_type
    if isinstance(data_type, EclTypeEnum):
        if not constant_size_data_type(data_type):
            raise ValueError("Cannot cast EclTypeEnum (%d) to EclDataType due "
                    "to non-constant size. Please provide an EclDataType instead.")

        dump_type_deprecation_warning()
        return EclDataType(data_type)

class EclKW(BaseCClass):
    """
    The EclKW class contains the information from one ECLIPSE keyword.

    The ecl_kw type is the lowest level type in the libecl C library,
    and all the other datatypes like e.g. ecl_grid and ecl_sum are
    based on collections of ecl_kw instances, and interpreting the
    content of the ecl_kw keywords.

    Many of the special __xxx___() functions have been implemented, so
    that the EclKW class supports both numerical operations and also
    [] based lookup. Many of the methods accept an optional @mask
    argument; this should be a EclRegion instance which can be used to
    limit the operation to a part of the EclKW.
    """

    int_kw_set = set(["PVTNUM", "FIPNUM", "EQLNUM", "FLUXNUM", "MULTNUM", "ACTNUM", "SPECGRID", "REGIONS"])

    TYPE_NAME          = "ecl_kw"
    _alloc_new         = EclPrototype("void* ecl_kw_alloc_python(char*, int, ecl_data_type)", bind = False)
    _fread_alloc       = EclPrototype("ecl_kw_obj ecl_kw_fread_alloc(fortio)", bind = False)
    _load_grdecl       = EclPrototype("ecl_kw_obj ecl_kw_fscanf_alloc_grdecl_dynamic_python(FILE, char*, bool, ecl_data_type)", bind = False)
    _fseek_grdecl      = EclPrototype("bool     ecl_kw_grdecl_fseek_kw(char*, bool, FILE)", bind = False)

    _sub_copy          = EclPrototype("ecl_kw_obj ecl_kw_alloc_sub_copy(ecl_kw, char*, int, int)")
    _copyc             = EclPrototype("ecl_kw_obj ecl_kw_alloc_copy(ecl_kw)")
    _slice_copyc       = EclPrototype("ecl_kw_obj ecl_kw_alloc_slice_copy(ecl_kw, int, int, int)")
    _global_copy       = EclPrototype("ecl_kw_obj ecl_kw_alloc_global_copy(ecl_kw, ecl_kw)")
    _fprintf_grdecl    = EclPrototype("void     ecl_kw_fprintf_grdecl(ecl_kw, FILE)")
    _fprintf_data      = EclPrototype("void     ecl_kw_fprintf_data(ecl_kw, char*, FILE)")

    _get_size          = EclPrototype("int      ecl_kw_get_size(ecl_kw)")
    _get_fortio_size   = EclPrototype("size_t   ecl_kw_fortio_size(ecl_kw)")
    _get_type          = EclPrototype("ecl_type_enum ecl_kw_get_type(ecl_kw)")
    _iget_char_ptr     = EclPrototype("char*    ecl_kw_iget_char_ptr(ecl_kw, int)")
    _iset_char_ptr     = EclPrototype("void     ecl_kw_iset_char_ptr(ecl_kw, int, char*)")
    _iget_string_ptr   = EclPrototype("char*    ecl_kw_iget_string_ptr(ecl_kw, int)")
    _iset_string_ptr   = EclPrototype("void     ecl_kw_iset_string_ptr(ecl_kw, int, char*)")
    _iget_bool         = EclPrototype("bool     ecl_kw_iget_bool(ecl_kw, int)")
    _iset_bool         = EclPrototype("bool     ecl_kw_iset_bool(ecl_kw, int, bool)")
    _iget_int          = EclPrototype("int      ecl_kw_iget_int(ecl_kw, int)")
    _iget_double       = EclPrototype("double   ecl_kw_iget_double(ecl_kw, int)")
    _iget_float        = EclPrototype("float    ecl_kw_iget_float(ecl_kw, int)")
    _float_ptr         = EclPrototype("float*   ecl_kw_get_float_ptr(ecl_kw)")
    _int_ptr           = EclPrototype("int*     ecl_kw_get_int_ptr(ecl_kw)")
    _double_ptr        = EclPrototype("double*  ecl_kw_get_double_ptr(ecl_kw)")
    _free              = EclPrototype("void     ecl_kw_free(ecl_kw)")
    _fwrite            = EclPrototype("void     ecl_kw_fwrite(ecl_kw, fortio)")
    _get_header        = EclPrototype("char*    ecl_kw_get_header (ecl_kw)")
    _set_header        = EclPrototype("void     ecl_kw_set_header_name (ecl_kw, char*)")
    _get_data_type     = EclPrototype("ecl_data_type_obj ecl_kw_get_data_type_python(ecl_kw)");

    _int_sum           = EclPrototype("int      ecl_kw_element_sum_int(ecl_kw)")
    _float_sum         = EclPrototype("double   ecl_kw_element_sum_float(ecl_kw)")
    _iadd_squared      = EclPrototype("void     ecl_kw_inplace_add_squared(ecl_kw, ecl_kw)")
    _isqrt             = EclPrototype("void     ecl_kw_inplace_sqrt(ecl_kw)")
    _iadd              = EclPrototype("void     ecl_kw_inplace_add(ecl_kw, ecl_kw)")
    _imul              = EclPrototype("void     ecl_kw_inplace_mul(ecl_kw, ecl_kw)")
    _idiv              = EclPrototype("void     ecl_kw_inplace_div(ecl_kw, ecl_kw)")
    _isub              = EclPrototype("void     ecl_kw_inplace_sub(ecl_kw, ecl_kw)")
    _iabs              = EclPrototype("void     ecl_kw_inplace_abs(ecl_kw)")
    _equal             = EclPrototype("bool     ecl_kw_equal(ecl_kw, ecl_kw)")
    _equal_numeric     = EclPrototype("bool     ecl_kw_numeric_equal(ecl_kw, ecl_kw, double, double)")

    _assert_binary     = EclPrototype("bool     ecl_kw_size_and_numeric_type_equal(ecl_kw, ecl_kw)")
    _scale_int         = EclPrototype("void     ecl_kw_scale_int(ecl_kw, int)")
    _scale_float       = EclPrototype("void     ecl_kw_scale_float_or_double(ecl_kw, double)")
    _shift_int         = EclPrototype("void     ecl_kw_shift_int(ecl_kw, int)")
    _shift_float       = EclPrototype("void     ecl_kw_shift_float_or_double(ecl_kw, double)")
    _copy_data         = EclPrototype("void     ecl_kw_memcpy_data(ecl_kw, ecl_kw)")
    _set_int           = EclPrototype("void     ecl_kw_scalar_set_int(ecl_kw, int)")
    _set_float         = EclPrototype("void     ecl_kw_scalar_set_float_or_double(ecl_kw, double)")

    _max_min_int       = EclPrototype("void     ecl_kw_max_min_int(ecl_kw, int*, int*)")
    _max_min_float     = EclPrototype("void     ecl_kw_max_min_float(ecl_kw, float*, float*)")
    _max_min_double    = EclPrototype("void     ecl_kw_max_min_double(ecl_kw, double*, double*)")
    _fix_uninitialized = EclPrototype("void     ecl_kw_fix_uninitialized(ecl_kw,int, int, int, int*)")
    _create_actnum     = EclPrototype("ecl_kw_obj ecl_kw_alloc_actnum(ecl_kw, float)")
    _first_different   = EclPrototype("int      ecl_kw_first_different(ecl_kw, ecl_kw, int, double, double)")
    _resize            = EclPrototype("void     ecl_kw_resize(ecl_kw, int)")
    _safe_div          = EclPrototype("bool     ecl_kw_inplace_safe_div(ecl_kw,ecl_kw)")


    @classmethod
    def createCReference(cls, c_ptr, parent=None):
        ecl_kw = super(EclKW, cls).createCReference(c_ptr, parent=parent)
        if ecl_kw is None:
            raise ValueError("Failed to create EclKW instance")

        ecl_kw.__private_init()
        return ecl_kw


    @classmethod
    def createPythonObject(cls, c_ptr):
        ecl_kw = super(EclKW, cls).createPythonObject(c_ptr)
        if ecl_kw is None:
            raise ValueError("Failed to create EclKW instance")

        ecl_kw.__private_init()
        return ecl_kw



    @classmethod
    def add_int_kw(cls, kw):
        """Will add keyword @kw to the standard set of integer keywords."""
        cls.int_kw_set.add(kw)

    @classmethod
    def del_int_kw(cls, kw):
        """Will remove keyword @kw from the standard set of integer keywords."""
        cls.int_kw_set.discard(kw)

    @classmethod
    def int_keywords(cls):
        """Will return the current set of integer keywords."""
        return cls.int_kw_set


    def slice_copy(self, slice_range):
        (start, stop, step) = slice_range.indices(len(self))
        if stop > start:
            return self._slice_copyc(start, stop, step)
        else:
            return None


    def copy(self):
        """
        Will create a deep copy of the current kw instance.
        """
        return self._copyc()




    @classmethod
    def read_grdecl(cls, fileH, kw, strict=True, ecl_type=None):
        """
        Function to load an EclKW instance from a grdecl formatted filehandle.

        This constructor can be used to load an EclKW instance from a
        grdecl formatted file; the input files for petrophysical
        properties are typically given as grdecl files.

        The @file argument should be a Python filehandle to an open
        file. The @kw argument should be the keyword header you are
        searching for, e.g. "PORO" or "PVTNUM"[1], the method will
        then search forward through the file to look for this @kw. If
        the keyword can not be found the method will return None. The
        searching will start from the current position in the file; so
        if you want to reposition the file pointer you should use the
        seek() method of the file object first.

        Observe that there is a strict 8 character limit on @kw -
        altough you could in principle use an arbitrary external
        program to create grdecl files with more than 8 character
        length headers, this implementation will refuse to even try
        loading them. In that case you will have to rename the
        keywords in your file - sorry. A TypeError exception
        will be raised if @kw has more than 8 characters.

        The implementation in ert can read integer and float type
        keywords from grdecl files; however the grdecl files have no
        datatype header, and it is impossible to determine the type
        reliably by inspection. Hence the type must be known when
        reading the file. The algorithm for specifying type, in order
        of presedence, is as follows:

        1. The optional argument @ecl_type can be used to specify
           the type:

           special_int_kw = EclKW.read_grdecl(fileH, 'INTKW', ecl_type=ECL_INT)

           If ecl_type is different from ECL_INT or
           ECL_FLOAT a TypeError exception will be raised.

           If ecl_type == None (the default), the method will continue
           to point 2. or 3. to determine the correct type.


        2. If the keyword is included in the set built in set
           'int_kw_set' the type will be ECL_INT_TYPE.

           pvtnum_kw = EclKW.read_grdecl(fileH, 'PVTNUM')

           Observe that (currently) no case conversions take place
           when checking the 'int_kw_set'. The current built in set is
           accesible through the int_kw property.


        3. Otherwise the default is float, i.e. ECL_FLOAT.

           EclKw reads grdecl with EclDataType
           poro_kw = EclKW.read_grdecl(fileH, 'PORO')


        Observe that since the grdecl files are quite weakly
        structured it is difficult to verify the integrity of the
        files, malformed input might therefore pass unnoticed before
        things blow up at a later stage.

        [1]: It is possible, but not recommended, to pass in None for
        @kw, in which case the method will load the first keyword
        it finds in the file.
        """

        cfile  = CFILE(fileH)
        if kw:
            if len(kw) > 8:
                raise TypeError("Sorry keyword:%s is too long, must be eight characters or less." % kw)

        if ecl_type is None:
            if cls.int_kw_set.__contains__(kw):
                ecl_type = EclDataType.ECL_INT
            else:
                ecl_type = EclDataType.ECL_FLOAT

        ecl_type = warn_and_cast_data_type(ecl_type)

        if not isinstance(ecl_type, EclDataType):
            raise TypeError("Expected EclDataType, was: %s" % type(ecl_type))

        if not ecl_type in [EclDataType.ECL_FLOAT, EclDataType.ECL_INT]:
            raise ValueError("The type:%s is invalid when loading keyword:%s" % (ecl_type.type_name, kw))

        return cls._load_grdecl(cfile, kw, strict, ecl_type)


    @classmethod
    def fseek_grdecl(cls, fileH, kw, rewind=False):
        """
        Will search through the open file and look for string @kw.

        If the search succeeds the function will return and the file
        pointer will be positioned at the start of the kw, if the
        search fails the function will return false and the file
        pointer will be repositioned at the position it had prior to
        the call.

        Only @kw instances which are found at the beginning of a line
        (with optional leading space characters) are considered,
        i.e. searching for the string PERMX in the cases below will
        fail:

           -- PERMX
           EQUIL   PERMX /


        The function will start searching from the current position in
        the file and forwards, if the optional argument @rewind is
        true the function rewind to the beginning of the file and
        search from there after the initial search.
        """
        cfile = CFILE(fileH)
        return cls._fseek_grdecl(kw, rewind, cfile)


    @classmethod
    def fread(cls, fortio):
        """
        Will read a new EclKW instance from the open FortIO file.
        """
        return cls._fread_alloc(fortio)


    def free(self):
        self._free()

    def __repr__(self):
        si = len(self)
        nm = self.getName()
        mm = 'type=%s' % str(self.data_type)
        if self.isNumeric():
            mi, ma = self.getMinMax()
            mm = 'min=%.2f, max=%.2f' % (mi,ma)
        ad = self._ad_str()
        fmt = 'EclKW(size=%d, name="%s", %s) %s'
        return fmt % (si,nm,mm,ad)

    def __init__(self, name, size, data_type):
        """Creates a brand new EclKW instance.

        This method will create a grand spanking new EclKW
        instance. The instance will get name @name @size elements and
        datatype @data_type. Using this method you could create a SOIL
        keyword with:

           soil_kw = EclKW("SOIL", 10000, ECL_FLOAT_TYPE)

        """
        if len(name) > 8:
            raise ValueError("Sorry - maximum eight characters in keyword name")

        data_type = warn_and_cast_data_type(data_type)

        if not isinstance(data_type, EclDataType):
            raise TypeError("Expected an EclDataType, received: %s" %
                    type(data_type))

        c_ptr = self._alloc_new(name, size, data_type)
        super(EclKW, self).__init__(c_ptr)
        self.__private_init()



    def __private_init(self):
        self.data_ptr   = None

        if self.data_type.is_int():
            self.data_ptr = self._int_ptr()
            self.dtype    = numpy.int32
            self.str_fmt  = "%8d"
        elif self.data_type.is_float():
            self.data_ptr = self._float_ptr()
            self.dtype    = numpy.float32
            self.str_fmt  = "%13.4f"
        elif self.data_type.is_double():
            self.data_ptr = self._double_ptr()
            self.dtype    = numpy.float64
            self.str_fmt  = "%13.4f"
        else:
            # Iteration not supported for CHAR / BOOL
            self.data_ptr = None
            self.dtype    = None
            if self.data_type.is_char():
                self.str_fmt  = "%8s"
            elif self.data_type.is_bool():
                self.str_fmt  = "%d"
            elif self.data_type.is_mess():
                self.str_fmt = "%s"  #"Message type"
            elif self.data_type.is_string():
                self.str_fmt = "%" + str(self.data_type.element_size) + "s"
            else:
                raise ValueError("Unknown EclDataType (%s)!" % self.data_type.type_name)

    def sub_copy(self, offset, count, new_header=None):
        """
        Will create a new block copy of the src keyword.

        If @new_header == None the copy will get the same 'name' as
        the src, otherwise the keyword will get the @new_header as
        header.

        The copy will start at @block of the src keyword and copy
        @count elements; a negative value of @count is interpreted as
        'the rest of the elements'

           new1 = src.sub_copy(0, 10, new_header="NEW1")
           new2 = src.sub_copy(10, -1, new_header="NEW2")

        If the count or index arguments are in some way invalid the
        method will raise IndexError.
        """
        if offset < 0 or offset >= len(self):
            raise IndexError("Offset:%d invalid - valid range:[0,%d)" % (offset, len(self)))

        if offset + count > len(self):
            raise IndexError("Invalid value of (offset + count):%d" % (offset + count))

        return self._sub_copy(new_header, offset, count)


    def is_numeric(self):
        """
        Will check if the keyword contains numeric data, i.e int, float or double.
        """
        return self.data_type.is_numeric()

    def ecl_kw_instance(self):
        return True



    def __len__(self):
        """
        Returns the number of elements. Implements len()
        """
        return self._get_size()


    def __deep_copy__(self, memo):
        """
        Python special routine used to perform deep copy.
        """
        ecl_kw = self.copy()
        return ecl_kw


    def __getitem__(self, index):
        """
        Function to support index based lookup: y = kw[index]
        """
        if isinstance(index,int):
            length = len(self)
            if index < 0:
                # We allow one level of negative indexing
                index += len(self)

            if index < 0 or index >= length:
                raise IndexError
            else:
                if self.data_ptr:
                    return self.data_ptr[ index ]
                else:
                    if self.data_type.is_bool():
                        return self._iget_bool(index)
                    elif self.data_type.is_char():
                        return self._iget_char_ptr(index)
                    elif self.data_type.is_string():
                        return self._iget_string_ptr(index)
                    else:
                        raise TypeError("Internal implementation error ...")
        elif isinstance(index, slice):
            return self.slice_copy(index)
        else:
            raise TypeError("Index should be integer type")


    def __setitem__(self, index,value):
        """
        Function to support index based assignment: kw[index] = value
        """
        if isinstance(index, int):
            length = len(self)
            if index < 0:
                # Will only wrap backwards once
                index = len(self) + index

            if index < 0 or index >= length:
                raise IndexError
            else:
                if self.data_ptr:
                    self.data_ptr[ index ] = value
                else:
                    if self.data_type.is_bool():
                        self._iset_bool(index, value)
                    elif self.data_type.is_char():
                        return self._iset_char_ptr(index, value)
                    elif self.data_type.is_string():
                        return self._iset_string_ptr(index, value)
                    else:
                        raise SystemError("Internal implementation error ...")
        elif isinstance(index, slice):
            (start, stop, step) = index.indices(len(self))
            index = start
            while index < stop:
                self[index] = value
                index += step
        else:
            raise TypeError("Index should be integer type")


    #################################################################


    def __IMUL__(self, factor, mul=True):
        if self.isNumeric():
            if hasattr(factor, "ecl_kw_instance"):
                if self.assert_binary(factor):
                    if mul:
                        self._imul(factor)
                    else:
                        self._idiv(factor)
                else:
                    raise TypeError("Type mismatch")
            else:
                if not mul:
                    factor = 1.0 / factor

                if self.data_type.is_int():
                    if isinstance(factor, int):
                        self._scale_int(factor)
                    else:
                        raise TypeError("Type mismatch")
                else:
                    if isinstance(factor, int) or isinstance(factor, float):
                        self._scale_float(factor)
                    else:
                        raise TypeError("Only muliplication with scalar supported")
        else:
            raise TypeError("Not numeric type")

        return self


    def __IADD__(self, delta, add=True):
        if self.isNumeric():
            if type(self) == type(delta):
                if self.assert_binary(delta):
                    if add:
                        self._iadd(delta)
                    else:
                        self._isub(delta)
                else:
                    raise TypeError("Type / size mismatch")
            else:
                if add:
                    sign = 1
                else:
                    sign = -1

                if self.data_type.is_int():
                    if isinstance(delta, int):
                        self._shift_int(delta * sign)
                    else:
                        raise TypeError("Type mismatch")
                else:
                    if isinstance(delta, int) or isinstance(delta, float):
                        self._shift_float(delta * sign) # Will call the _float() or _double() function in the C layer.
                    else:
                        raise TypeError("Type mismatch")
        else:
            raise TypeError("Type / size mismatch")

        return self

    def __iadd__(self, delta):
        return self.__IADD__(delta, True)

    def __isub__(self, delta):
        return self.__IADD__(delta, False)

    def __imul__(self, delta):
        return self.__IMUL__(delta, True)

    def __idiv__(self, delta):
        return self.__IMUL__(delta, False)


    #################################################################

    def __abs__(self):
        if self.isNumeric():
            copy = self.copy()
            copy._iabs()
            return copy
        else:
            raise TypeError("The __abs__() function is only implemented for numeric types")



    def __add__(self, delta):
        copy = self.copy()
        copy += delta
        return copy

    def __radd__(self, delta):
        return self.__add__(delta)

    def __sub__(self, delta):
        copy  = self.copy()
        copy -= delta
        return copy

    def __rsub__(self, delta):
        return self.__sub__(delta) * -1

    def __mul__(self, factor):
        copy  = self.copy()
        copy *= factor
        return copy

    def __rmul__(self, factor):
        return self.__mul__(factor)

    def __div__(self, factor):
        copy = self.copy()
        copy /= factor
        return copy

    # No __rdiv__()

    def add_squared(self, other):
       if not self.is_numeric():
           raise TypeError("Can only be called on numeric types")

       if not self.assert_binary(other):
           raise ValueError("Invalid argument to method add_squared")

       self._iadd_squared(other)

    def isqrt(self):
        if not self.is_numeric():
            raise TypeError("Can only be called on numeric types")

        self._isqrt()


    def sum(self, mask = None, force_active = False):
        """
        Will calculate the sum of all the elements in the keyword.

        String: Raise ValueError exception.
        Bool:   The number of true values
        """
        if mask is None:
            if self.data_type.is_int():
                return self._int_sum()
            elif self.data_type.is_float():
                return self._float_sum()
            elif self.data_type.is_double():
                return self._float_sum()
            elif self.data_type.is_bool():
                sum = 0
                for elm in self:
                    if elm:
                        sum += 1
                return sum
            else:
                raise ValueError('The keyword "%s" is of string type - sum is not implemented' % self.getName())

        return mask.sum_kw(self, force_active)



    def assert_binary(self, other):
        """
        Utility function to assert that keywords @self and @other can
        be combined.
        """
        return self._assert_binary(other)

    #################################################################

    def assign(self, value, mask=None, force_active=False):
        """
        Assign a value to current kw instance.

        This method is used to assign value(s) to the current EclKW
        instance. The @value parameter can either be a scalar, or
        another EclKW instance. To set all elements of a keyword to
        1.0:

            kw.assign(1.0)

        The numerical type of @value must be compatible with the
        current keyword. The optional @mask argument should be an
        EclRegion instance which can be used to limit the assignment
        to only parts of the EclKW. In the example below we select all
        the elements with PORO below 0.10, and then assign EQLNUM
        value 88 to those cells:

            grid = ecl.EclGrid("ECLIPSE.EGRID")
            reg  = ecl.EclRegion(grid, false)
            init = ecl.EclFile("ECLIPSE.INIT")

            poro = init["PORO"][0]
            eqlnum = init["EQLNUM"][0]
            reg.select_below(poro, 0.10)

            eqlnum.assign(88, mask = reg)

        The EclRegion instance has two equivalent sets of selected
        indices; one consisting of active indices and one consisting
        of global indices. By default the assign() method will select
        the global indices if the keyword has nx*ny*nz elements and
        the active indices if the kw has nactive elements. By setting
        the optional argument @force_active to true, you can force the
        method to only modify the active indices, even though the
        keyword has nx*ny*nz elements; if the keyword has nactive
        elements the @force_active flag is not considered.
        """
        if self.isNumeric():
            if type(value) == type(self):
                if mask is not None:
                    mask.copy_kw(self, value, force_active)
                else:
                    if self.assert_binary(value):
                        self._copy_data(value)
                    else:
                        raise TypeError("Type / size mismatch")
            else:
                if mask is not None:
                    mask.set_kw(self, value, force_active)
                else:
                    if self.data_type.is_int():
                        if isinstance(value, int):
                            self._set_int(value)
                        else:
                            raise TypeError("Type mismatch")
                    else:
                        if isinstance(value, int) or isinstance(value, float):
                            self._set_float(value)
                        else:
                            raise TypeError("Only muliplication with scalar supported")


    def add(self, other, mask=None, force_active=False):
        """
        See method assign() for documentation of optional arguments
        @mask and @force_active.
        """
        if mask is not None:
            mask.iadd_kw(self, other, force_active)
        else:
            return self.__iadd__(other)

    def sub(self, other, mask=None, force_active=False):
        """
        See method assign() for documentation of optional arguments
        @mask and @force_active.
        """
        if mask is not None:
            mask.isub_kw(self, other, force_active)
        else:
            return self.__isub__(other)

    def mul(self, other, mask=None, force_active=False):
        """
        See method assign() for documentation of optional arguments
        @mask and @force_active.
        """
        if mask is not None:
            mask.imul_kw(self, other, force_active)
        else:
            return self.__imul__(other)

    def div(self, other, mask=None, force_active=False):
        """
        See method assign() for documentation of optional arguments
        @mask and @force_active.
        """
        if mask is not None:
            mask.idiv_kw(self, other, force_active)
        else:
            return self.__idiv__(other)

    def apply(self, func, arg=None, mask=None, force_active=False):
        """
        Will apply the function @func on the keyword - inplace.

        The function @func should take a scalar value from the ecl_kw
        vector as input, and return a scalar value of the same type;
        optionally you can supply a second argument with the @arg
        attribute:

          def cutoff(x, limit):
              if x > limit:
                 return x
              else:
                 return 0


          kw.apply(math.sin)
          kw.apply(cutoff, arg=0.10)


        See method assign() for documentation of optional arguments
        @mask and @force_active.
        """
        if mask is not None:
            active_list = mask.kw_index_list(self, force_active)
            if arg:
                for index in active_list:
                    self.data_ptr[index] = func(self.data_ptr[index], arg)
            else:
                for index in active_list:
                    self.data_ptr[index] = func(self.data_ptr[index])
        else:
            if arg:
                for i in range(len(self)):
                    self.data_ptr[i] = func(self.data_ptr[i], arg)
            else:
                for i in range(len(self)):
                    self.data_ptr[i] = func(self.data_ptr[i])


    def equal(self, other):
        """
        Will check if the two keywords are (exactly) equal.

        The check is based on the content of the keywords, and not
        pointer comparison.
        """
        if isinstance(other, EclKW):
            return self._equal(other)
        else:
            raise TypeError("Can only compare with another EclKW")


    def __eq__(self, other):
        return self.equal(other)

    def __hash__(self):
        return hash(self._get_header())


    def equal_numeric(self, other, epsilon=1e-6, abs_epsilon=None, rel_epsilon=None):
        """Will check if two numerical keywords are ~nearly equal.


        If the keywords are of type integer, the comparison is
        absolute.

        If you pass in xxx_epsilon <= 0 the xxx_epsilon will be
        ignored in the test.

        """
        if isinstance(other, EclKW):
            if abs_epsilon is None:
                abs_epsilon = epsilon

            if rel_epsilon is None:
                rel_epsilon = epsilon

            return self._equal_numeric(other, abs_epsilon, rel_epsilon)
        else:
            raise TypeError("Can only compare with another EclKW")


    #################################################################

    def deep_copy(self):
        ecl_kw = self.__deep_copy__({})
        return ecl_kw

    def fort_io_size(self):
        """
        The number of bytes this keyword would occupy in a BINARY file.
        """
        return self._get_fortio_size()

    def set_name(self, name):
        if len(name) > 8:
            raise ValueError("Sorry: the name property must be max 8 characters long :-(")
        self._set_header(name)

    @property
    def name(self):
        n = self._get_header()
        return str(n) if n else ''

    @name.setter
    def name(self, value):
        self.set_name(value)

    def get_name(self):
        return self.name

    def resize(self, new_size):
        """
        Will set the new size of the kw to @new_size.
        """
        if new_size >= 0:
            self._resize(int(new_size))

        # Iteration is based on a pointer to the underlying storage,
        # that will generally by reset by the resize() call; i.e. we
        # need to call the __private_init() method again.
        self.__private_init()


    def get_min_max(self):
        """
        Will return a touple (min,max) for numerical types.

        Will raise TypeError exception if the keyword is not of
        numerical type.
        """
        if self.data_type.is_float():
            min_ = ctypes.c_float()
            max_ = ctypes.c_float()
            self._max_min_float(ctypes.byref(max_), ctypes.byref(min_))
        elif self.data_type.is_double():
            min_ = ctypes.c_double()
            max_ = ctypes.c_double()
            self._max_min_double(ctypes.byref(max_), ctypes.byref(min_))
        elif self.data_type.is_int():
            min_ = ctypes.c_int()
            max_ = ctypes.c_int()
            self._max_min_int(ctypes.byref(max_), ctypes.byref(min_))
        else:
            raise TypeError("min_max property not defined for keywords of type: %s" % self.type)
        return (min_.value, max_.value)


    def get_max(self):
        mm = self.getMinMax()
        return mm[1]


    def get_min(self):
        mm = self.getMinMax()
        return mm[0]

    @property
    def type(self):
        warnings.warn("ecl_kw.type is deprecated, use .data_type",
            DeprecationWarning)
        return self._get_type()

    @property
    def data_type(self):
        return self._get_data_type()

    @property
    def type_name(self):
        return self.data_type.type_name

    def type_name(self):
        return self.data_type.type_name

    def get_ecl_type(self):
        warnings.warn("EclTypeEnum is deprecated. " +
            "You should instead provide an EclDataType",
            DeprecationWarning)

        return self._get_type()




    @property
    def header(self):
        return (self.getName(), len(self), self.typeName())

    @property
    def array(self):
        a = self.data_ptr
        if not a == None:
            a.size        = len(self)
            a.__parent__  = self  # Inhibit GC
        return a


    def str_data(self, width, index1, index2, fmt):
        """
        Helper function for str() method.
        """
        data = []
        s = ""
        for index in range(index1, index2):
            data.append(self[index])
        for index in range(len(data)):
            s += fmt % data[ index ]
            if index % width == (width - 1):
                s+= "\n"
        return s


    def str(self, width=5, max_lines=10, fmt=None):
        """
        Return string representation of kw for pretty printing.

        The function will return a string consisting of a header, and
        then a chunk of data. The data will be formatted in @width
        columns, and a maximum of @max_lines lines. If @max_lines is
        not sufficient the first elements in the kewyord are
        represented, a .... continuation line and then the last part
        of the keyword. If @max_lines is None all of the vector will
        be printed, irrespective of how long it is.

        If a value is given for @fmt that is used as format string for
        each element, otherwise a type-specific default format is
        used. If given the @fmt string should contain spacing between
        the elements. The implementation of the builtin method
        __str__() is based on this method.
        """
        s = "%-8s %8d %-4s\n" % (self.getName(), len(self), self.typeName())
        lines = len(self) // width
        if not fmt:
            fmt = self.str_fmt + " "

        if max_lines is None or lines <= max_lines:
            s += self.str_data(width, 0, len(self), fmt)
        else:
            s1 = width * max_lines // 2
            s += self.str_data(width , 0, s1, fmt)
            s += "   ....   \n"
            s += self.str_data(width , len(self) - s1, len(self), fmt)

        return s

    def __str__(self):
        """
        Return string representation - see method str().
        """
        return self.str(width=5, max_lines=10)


    def numpy_view(self):
        """Will return a numpy view to the underlying data.

        The data in this numpy array is *shared* with the EclKW
        instance, meaning that updates in one will be reflected in the
        other.
        """

        if self.dtype is numpy.float64:
            ct = ctypes.c_double
        elif self.dtype is numpy.float32:
            ct = ctypes.c_float
        elif self.dtype is numpy.int32:
            ct = ctypes.c_int
        else:
            raise ValueError("Invalid type - numpy array only valid for int/float/double")

        ap = ctypes.cast(self.data_ptr, ctypes.POINTER(ct * len(self)))
        return numpy.frombuffer(ap.contents, dtype=self.dtype)


    def numpy_copy(self):
        """Will return a numpy array which contains a copy of the EclKW data.

        The numpy array has a separate copy of the data, so that
        changes to either the numpy array or the EclKW will *not* be
        reflected in the other datastructure. This is in contrast to
        the EclKW.numpyView() method where the underlying data is
        shared.
        """
        view = self.numpyView()
        return numpy.copy(view)

    def fwrite(self, fortio):
        self._fwrite(fortio)

    def write_grdecl(self, file):
        """
        Will write keyword in GRDECL format.

        This method will write the current keyword in GRDECL format,
        the @file argument must be a Python file handle to an already
        opened file. In the example below we load the porosity from an
        existing GRDECL file, set all poro values below 0.05 to 0.00
        and write back an updated GRDECL file.

            poro = ecl.EclKW.read_grdecl(open("poro1.grdecl", "r"), "PORO")
            grid = ecl.EclGrid("ECLIPSE.EGRID")
            reg  = ecl.EclRegion(grid, False)

            reg.select_below(poro, 0.05)
            poro.assign(0.0, mask=reg)

            fileH = open("poro2.grdecl", "w")
            poro.write_grdecl(fileH)
            fileH.close()

        """
        cfile = CFILE(file)
        self._fprintf_grdecl(cfile)



    def fprintf_data(self, file, fmt=None):
        """
        Will print the keyword data formatted to file.

        The @file argument should be a python file handle to a file
        opened for writing. The @fmt argument is used as fprintf()
        format specifier, observe that the format specifier should
        include a separation character between the elements. If no
        @fmt argument is supplied the default str_fmt specifier is
        used for every element, separated by a newline.

        In the case of boolean data the function will print o and 1
        for False and True respectively. For string data the function
        will print the data as 8 characters long string with blank
        padding on the right.
        """
        if fmt is None:
            fmt = self.str_fmt + "\n"
        cfile = CFILE(file)
        self._fprintf_data(fmt, cfile)

    def create_actnum(self, porv_limit = 0):
        """Will create ACTNUM keyword from PORV keyword.

        This quite specialized method will create an ACTNUM keyword based on
        interpreting the current keyword as a PORV keyword. The method will
        raise an exception if the current keyword is not ("PORV", FLOAT). The
        code implemented in C for speed is essentially:

             actnum = [ 1 if x > porv_limit else 0 for x in self ]

        """
        if not self.data_type.is_float():
            raise TypeError("The PORV keyword must be of type FLOAT")

        if not self.get_name() == "PORV":
            raise ValueError("Input argument must be PORV keyword")

        return self._create_actnum(porv_limit)


    def fix_uninitialized(self, grid):
        """
        Special case function for region code.
        """
        dims = grid.getDims()
        actnum = grid.exportACTNUM()
        self._fix_uninitialized(dims[0], dims[1], dims[2], actnum.getDataPtr())


    def get_data_ptr(self):
        if self.data_type.is_int():
            return self._int_ptr()
        elif self.data_type.is_float():
            return self._float_ptr()
        elif self.data_type.is_double():
            return self._double_ptr()
        else:
            raise ValueError("Only numeric types can export data pointer")


    def first_different(self, other, offset=0, epsilon=0, abs_epsilon=None, rel_epsilon=None):
        if len(self) != len(other):
            raise ValueError("Keywords must have equal size")

        if offset >= len(self):
            raise IndexError("Offset:%d invalid - size:%d" % (offset, len(self)))

        if self.data_type!= other.data_type:
            raise TypeError("The two keywords have different type")

        if abs_epsilon is None:
            abs_epsilon = epsilon

        if rel_epsilon is None:
            rel_epsilon = epsilon

        return self._first_different(other, offset, abs_epsilon, rel_epsilon)




    def scatter_copy(self, actnum):
        if not isinstance(actnum, EclKW):
            raise TypeError("The actnum argument must be of type EclKW")

        return self._global_copy(actnum)


    def safe_div(self, divisor):
        if not len(self) == len(divisor):
            raise ValueError("Length mismatch between %s and %s" % (self.name, divisor.name))

        if not self.is_numeric():
            raise TypeError("The self keyword must be of numeric type")

        if not divisor.is_numeric():
            raise TypeError("Must divide by numeric keyword")

        ok = self._safe_div( divisor )
        if not ok:
            raise NotImplementedError("safe_div not implemented for this type combination")




monkey_the_camel(EclKW, 'intKeywords', EclKW.int_keywords, classmethod)
monkey_the_camel(EclKW, 'isNumeric', EclKW.is_numeric)
monkey_the_camel(EclKW, 'fortIOSize', EclKW.fort_io_size)
monkey_the_camel(EclKW, 'setName', EclKW.set_name)
monkey_the_camel(EclKW, 'getName', EclKW.get_name)
monkey_the_camel(EclKW, 'getMinMax', EclKW.get_min_max)
monkey_the_camel(EclKW, 'getMax', EclKW.get_max)
monkey_the_camel(EclKW, 'getMin', EclKW.get_min)
monkey_the_camel(EclKW, 'typeName', EclKW.type_name)
monkey_the_camel(EclKW, 'getEclType', EclKW.get_ecl_type)
monkey_the_camel(EclKW, 'numpyView', EclKW.numpy_view)
monkey_the_camel(EclKW, 'numpyCopy', EclKW.numpy_copy)
monkey_the_camel(EclKW, 'fixUninitialized', EclKW.fix_uninitialized)
monkey_the_camel(EclKW, 'getDataPtr', EclKW.get_data_ptr)
monkey_the_camel(EclKW, 'firstDifferent', EclKW.first_different)



#  Copyright (C) 2018  Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

"""
The eclfile package contains several classes for working directly with ECLIPSE
files. 

  fortio/FortIO: This is functionality to read and write binary
     fortran files.

  ecl_kw/EclKW: This class holds one ECLIPSE keyword, like SWAT, in
     restart format.

  ecl_type/EclDataType: This class is used to represent the data type
    of the elements in EclKW.

  ecl_file/EclFile: This class is used to load an ECLIPSE file in
     restart format, alternatively only parts of the file can be
     loaded. Internally it consists of a collection of EclKW
     instances.
"""

import ecl.util.util

from .fortio import FortIO, openFortIO
from .ecl_kw import EclKW
from .ecl_file_view import EclFileView
from .ecl_file import EclFile , openEclFile
from .ecl_3dkw import Ecl3DKW
from .ecl_3d_file import Ecl3DFile
from .ecl_init_file import EclInitFile
from .ecl_restart_file import EclRestartFile



from ecl.well import WellTypeEnum
from ecl.well import WellConnectionDirectionEnum
from ecl.well import WellConnection
from ecl.well import WellSegment
from ecl.well import WellState
from ecl.well import WellTimeLine
from ecl.well import WellInfo



#  Copyright (C) 2011  Statoil ASA, Norway.
#
#  The file 'ecl_sum_keyword_vector.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.


import numpy
import datetime

# Observe that there is some convention conflict with the C code
# regarding order of arguments: The C code generally takes the time
# index as the first argument and the key/key_index as second
# argument. In the python code this order has been reversed.

from cwrap import BaseCClass
from ecl.util.util import monkey_the_camel
from ecl import EclPrototype



class EclSumKeyWordVector(BaseCClass):
    TYPE_NAME       = "ecl_sum_vector"
    _alloc          = EclPrototype("void* ecl_sum_vector_alloc(ecl_sum, bool)", bind=False)
    _alloc_copy     = EclPrototype("ecl_sum_vector_obj ecl_sum_vector_alloc_layout_copy(ecl_sum_vector, ecl_sum)")
    _free           = EclPrototype("void ecl_sum_vector_free(ecl_sum_vector)")
    _add            = EclPrototype("bool ecl_sum_vector_add_key(ecl_sum_vector,  char*)")
    _add_multiple   = EclPrototype("void ecl_sum_vector_add_keys(ecl_sum_vector,  char*)")
    _get_size       = EclPrototype("int ecl_sum_vector_get_size(ecl_sum_vector)")
    _iget_key       = EclPrototype("char* ecl_sum_vector_iget_key(ecl_sum_vector, int)")

    def __init__(self, ecl_sum, add_keywords = False):
        c_pointer = self._alloc(ecl_sum, add_keywords)
        super(EclSumKeyWordVector, self).__init__(c_pointer)

    def __getitem__(self, index):
        if index < 0:
            index += len(self)

        if index >= len(self):
            raise IndexError("Out of range")

        return self._iget_key( index )

    def __len__(self):
        return self._get_size()

    def free(self):
        self._free()

    def add_keyword(self, keyword):
        success = self._add(keyword)
        if not success:
            raise KeyError("Failed to add keyword to vector")

    def add_keywords(self, keyword_pattern):
        self._add_multiple(keyword_pattern)

    def __repr__(self):
        return self._create_repr('len=%d' % len(self))

    def copy(self, ecl_sum):
        return self._alloc_copy(ecl_sum)


monkey_the_camel(EclSumKeyWordVector, 'addKeyword', EclSumKeyWordVector.add_keyword)
monkey_the_camel(EclSumKeyWordVector, 'addKeywords', EclSumKeyWordVector.add_keywords)



#  Copyright (C) 2014  Statoil ASA, Norway.
#
#  The file 'double_vector.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from ecl import EclPrototype
from ecl.util.util import VectorTemplate


class DoubleVector(VectorTemplate):
    default_format = "%8.4f"

    _alloc            = EclPrototype("void*  double_vector_alloc( int , double )" , bind = False)
    _alloc_copy       = EclPrototype("double_vector_obj   double_vector_alloc_copy( double_vector )")
    _strided_copy     = EclPrototype("double_vector_obj   double_vector_alloc_strided_copy( double_vector , int , int , int)")
    _free             = EclPrototype("void   double_vector_free( double_vector )")
    _iget             = EclPrototype("double double_vector_iget( double_vector , int )")
    _safe_iget        = EclPrototype("double double_vector_safe_iget(double_vector , int )")
    _iset             = EclPrototype("double double_vector_iset( double_vector , int , double)")
    _size             = EclPrototype("int    double_vector_size( double_vector )")
    _append           = EclPrototype("void   double_vector_append( double_vector , double )")
    _idel_block       = EclPrototype("void   double_vector_idel_block( double_vector , int , int )")
    _pop              = EclPrototype("double double_vector_pop( double_vector )")
    _idel             = EclPrototype("void   double_vector_idel( double_vector , int )")
    _lshift           = EclPrototype("void   double_vector_lshift( double_vector , int )")
    _rshift           = EclPrototype("void   double_vector_rshift( double_vector , int )")
    _insert           = EclPrototype("void   double_vector_insert( double_vector , int , double)")
    _fprintf          = EclPrototype("void   double_vector_fprintf( double_vector , FILE , char* , char*)")
    _sort             = EclPrototype("void   double_vector_sort( double_vector )")
    _rsort            = EclPrototype("void   double_vector_rsort( double_vector )")
    _reset            = EclPrototype("void   double_vector_reset( double_vector )")
    _get_read_only    = EclPrototype("bool   double_vector_get_read_only( double_vector )")
    _set_read_only    = EclPrototype("void   double_vector_set_read_only( double_vector , bool )")
    _get_max          = EclPrototype("double double_vector_get_max( double_vector )")
    _get_min          = EclPrototype("double double_vector_get_min( double_vector )")
    _get_max_index    = EclPrototype("int    double_vector_get_max_index( double_vector , bool)")
    _get_min_index    = EclPrototype("int    double_vector_get_min_index( double_vector , bool)")
    _shift            = EclPrototype("void   double_vector_shift( double_vector , double )")
    _scale            = EclPrototype("void   double_vector_scale( double_vector , double )")
    _div              = EclPrototype("void   double_vector_div( double_vector , double )")
    _inplace_add      = EclPrototype("void   double_vector_inplace_add( double_vector , double_vector )")
    _inplace_mul      = EclPrototype("void   double_vector_inplace_mul( double_vector , double_vector )")
    _assign           = EclPrototype("void   double_vector_set_all( double_vector , double)")
    _memcpy           = EclPrototype("void   double_vector_memcpy(double_vector , double_vector )")
    _set_default      = EclPrototype("void   double_vector_set_default( double_vector , double)")
    _get_default      = EclPrototype("double double_vector_get_default( double_vector )")
    _element_size     = EclPrototype("int    double_vector_element_size( double_vector )")

    _permute          = EclPrototype("void double_vector_permute(double_vector, permutation_vector)")
    _sort_perm        = EclPrototype("permutation_vector_obj double_vector_alloc_sort_perm(double_vector)")
    _rsort_perm       = EclPrototype("permutation_vector_obj double_vector_alloc_rsort_perm(double_vector)")
    _contains         = EclPrototype("bool double_vector_contains(double_vector, double)")
    _select_unique    = EclPrototype("void double_vector_select_unique(double_vector)")
    _element_sum      = EclPrototype("double double_vector_sum(double_vector)")
    _get_data_ptr     = EclPrototype("double* double_vector_get_ptr(double_vector)")
    _count_equal      = EclPrototype("int double_vector_count_equal(double_vector, double)")
    _init_range       = EclPrototype("void double_vector_init_range(double_vector, double , double , double)")
    _init_linear      = EclPrototype("bool double_vector_init_linear(double_vector, double, double, int)")
    _equal            = EclPrototype("bool double_vector_equal(double_vector, double_vector)")
    _first_eq         = EclPrototype("int double_vector_first_equal(double_vector, double_vector, int)")
    _first_neq        = EclPrototype("int double_vector_first_not_equal(double_vector, double_vector, int)")



    def __init__(self, default_value=0, initial_size=0):
        super(DoubleVector, self).__init__(default_value, initial_size)



#  Copyright (C) 2014  Statoil ASA, Norway.
#
#  The file 'vector_template.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
import datetime
import re

from ecl import EclPrototype
from ecl.util.util import VectorTemplate, CTime


class TimeVector(VectorTemplate):
    TYPE_NAME = "time_t_vector"
    default_format = "%d"

    _alloc               = EclPrototype("void*   time_t_vector_alloc(int, time_t )" , bind = False)
    _alloc_copy          = EclPrototype("time_t_vector_obj time_t_vector_alloc_copy(time_t_vector )")
    _strided_copy        = EclPrototype("time_t_vector_obj time_t_vector_alloc_strided_copy(time_t_vector , int , int , int)")
    _free                = EclPrototype("void   time_t_vector_free( time_t_vector )")
    _iget                = EclPrototype("time_t   time_t_vector_iget( time_t_vector , int )")
    _safe_iget           = EclPrototype("time_t   time_t_vector_safe_iget( time_t_vector , int )")
    _iset                = EclPrototype("time_t   time_t_vector_iset( time_t_vector , int , time_t)")
    _size                = EclPrototype("int      time_t_vector_size( time_t_vector )")
    _append              = EclPrototype("void     time_t_vector_append( time_t_vector , time_t )")
    _idel_block          = EclPrototype("void     time_t_vector_idel_block( time_t_vector , int , int )")
    _idel                = EclPrototype("void     time_t_vector_idel( time_t_vector , int )")
    _pop                 = EclPrototype("time_t   time_t_vector_pop( time_t_vector )")
    _lshift              = EclPrototype("void     time_t_vector_lshift( time_t_vector , int )")
    _rshift              = EclPrototype("void     time_t_vector_rshift( time_t_vector , int )")
    _insert              = EclPrototype("void     time_t_vector_insert( time_t_vector , int , time_t)")
    _fprintf             = EclPrototype("void     time_t_vector_fprintf( time_t_vector , FILE , char* , char*)")
    _sort                = EclPrototype("void     time_t_vector_sort( time_t_vector )")
    _rsort               = EclPrototype("void     time_t_vector_rsort( time_t_vector )")
    _reset               = EclPrototype("void     time_t_vector_reset( time_t_vector )")
    _set_read_only       = EclPrototype("void     time_t_vector_set_read_only( time_t_vector , bool )")
    _get_read_only       = EclPrototype("bool     time_t_vector_get_read_only( time_t_vector )")
    _get_max             = EclPrototype("time_t   time_t_vector_get_max( time_t_vector )")
    _get_min             = EclPrototype("time_t   time_t_vector_get_min( time_t_vector )")
    _get_max_index       = EclPrototype("int      time_t_vector_get_max_index( time_t_vector , bool)")
    _get_min_index       = EclPrototype("int      time_t_vector_get_min_index( time_t_vector , bool)")
    _shift               = EclPrototype("void     time_t_vector_shift( time_t_vector , time_t )")
    _scale               = EclPrototype("void     time_t_vector_scale( time_t_vector , time_t )")
    _div                 = EclPrototype("void     time_t_vector_div( time_t_vector , time_t )")
    _inplace_add         = EclPrototype("void     time_t_vector_inplace_add( time_t_vector , time_t_vector )")
    _inplace_mul         = EclPrototype("void     time_t_vector_inplace_mul( time_t_vector , time_t_vector )")
    _assign              = EclPrototype("void     time_t_vector_set_all( time_t_vector , time_t)")
    _memcpy              = EclPrototype("void     time_t_vector_memcpy(time_t_vector , time_t_vector )")
    _set_default         = EclPrototype("void     time_t_vector_set_default( time_t_vector , time_t)")
    _get_default         = EclPrototype("time_t   time_t_vector_get_default( time_t_vector )")
    _element_size        = EclPrototype("int      time_t_vector_element_size( time_t_vector )")

    _permute             = EclPrototype("void time_t_vector_permute(time_t_vector, permutation_vector)")
    _sort_perm           = EclPrototype("permutation_vector_obj time_t_vector_alloc_sort_perm(time_t_vector)")
    _rsort_perm          = EclPrototype("permutation_vector_obj time_t_vector_alloc_rsort_perm(time_t_vector)")
    _contains            = EclPrototype("bool time_t_vector_contains(time_t_vector, time_t)")
    _select_unique       = EclPrototype("void time_t_vector_select_unique(time_t_vector)")
    _element_sum         = EclPrototype("time_t time_t_vector_sum(time_t_vector)")
    _count_equal         = EclPrototype("int time_t_vector_count_equal(time_t_vector, time_t)")
    _init_range          = EclPrototype("void time_t_vector_init_range(time_t_vector, time_t , time_t , time_t)")
    _init_linear         = EclPrototype("bool time_t_vector_init_linear(time_t_vector, time_t, time_t, int)")
    _equal               = EclPrototype("bool time_t_vector_equal(time_t_vector, time_t_vector)")
    _first_eq            = EclPrototype("int time_t_vector_first_equal(time_t_vector, time_t_vector, int)")
    _first_neq           = EclPrototype("int time_t_vector_first_not_equal(time_t_vector, time_t_vector, int)")



    def __init__(self, default_value=None, initial_size=0):
        if default_value is None:
            super(TimeVector, self).__init__(CTime(0), initial_size)
        else:
            try:
                default = CTime(default_value)
            except:
                raise ValueError("default value invalid - must be type ctime() or date/datetime")

            super(TimeVector, self).__init__(default, initial_size)

    @classmethod
    def parseTimeUnit(cls, deltaString):
        deltaRegexp = re.compile("(?P<num>\d*)(?P<unit>[dmy])", re.IGNORECASE)
        matchObj = deltaRegexp.match(deltaString)
        if matchObj:
            try:
                num = int(matchObj.group("num"))
            except:
                num = 1

            timeUnit = matchObj.group("unit").lower()
            return num, timeUnit
        else:
            raise TypeError("The delta string must be on form \'1d\', \'2m\', \'Y\' for one day, two months or one year respectively")

    def __str__(self):
        """
        Returns string representantion of vector.
        """
        string_list = []
        for d in self:
            string_list.append("%s" % d)

        return str(string_list)

    def append(self, value):
        self._append(CTime(value))

    def __contains__(self, value):
        return self._contains(CTime(value))

    def nextTime(self, num, timeUnit):
        currentTime = self[-1].datetime()
        hour = currentTime.hour
        minute = currentTime.minute
        second = currentTime.second

        if timeUnit == "d":
            td = datetime.timedelta(days=num)
            currentTime += td
        else:
            day = currentTime.day
            month = currentTime.month
            year = currentTime.year

            if timeUnit == "y":
                year += num
            else:
                month += num - 1
                (deltaYear, newMonth) = divmod(month, 12)
                month = newMonth + 1
                year += deltaYear
            currentTime = datetime.datetime(year, month, day, hour, minute, second)

        return currentTime

    def appendTime(self, num, timeUnit):
        next = self.nextTime(num, timeUnit)
        self.append(CTime(next))

    @classmethod
    def createRegular(cls, start, end, deltaString):
        """
        The last element in the vector will be <= end; i.e. if the
        question of whether the range is closed in the upper end
        depends on the commensurability of the [start,end] interval
        and the delta:

        createRegular(0 , 10 , delta=3) => [0,3,6,9]
        createRegular(0 , 10 , delta=2) => [0,2,4,6,8,10]
        """
        start = CTime(start)
        end = CTime(end)
        if start > end:
            raise ValueError("The time interval is invalid start is after end")

        (num, timeUnit) = cls.parseTimeUnit(deltaString)

        timeVector = TimeVector()
        currentTime = start
        while currentTime <= end:
            ct = CTime(currentTime)
            timeVector.append(ct)
            currentTime = timeVector.nextTime(num, timeUnit)

        return timeVector

    def getDataPtr(self):
        raise NotImplementedError("The getDataPtr() function is not implemented for time_t vectors")



#  Copyright (C) 2015  Statoil ASA, Norway. 
#   
#  The file 'arg_pack.py' is part of ERT - Ensemble based Reservoir Tool. 
#   
#  ERT is free software: you can redistribute it and/or modify 
#  it under the terms of the GNU General Public License as published by 
#  the Free Software Foundation, either version 3 of the License, or 
#  (at your option) any later version. 
#   
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY 
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or 
#  FITNESS FOR A PARTICULAR PURPOSE.   
#   
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html> 
#  for more details. 

from cwrap import BaseCClass
from ecl import EclPrototype


class ArgPack(BaseCClass):
    TYPE_NAME = "arg_pack"

    _alloc = EclPrototype("void* arg_pack_alloc()" , bind = False)
    _append_int = EclPrototype("void arg_pack_append_int(arg_pack, int)")
    _append_double = EclPrototype("void arg_pack_append_double(arg_pack, double)")
    _append_ptr = EclPrototype("void arg_pack_append_ptr(arg_pack, void*)")

    _size = EclPrototype("int arg_pack_size(arg_pack)")
    _free = EclPrototype("void arg_pack_free(arg_pack)")

    def __init__(self, *args):
        c_ptr = self._alloc()
        super(ArgPack, self).__init__(c_ptr)
        self.child_list = []
        for arg in args:
            self.append(arg)

            
    def append(self, data):
        if isinstance(data, int):
            self._append_int(data)
        elif isinstance(data, float):
            self._append_double(data)
        elif isinstance(data, BaseCClass):
            self._append_ptr(BaseCClass.from_param(data))
            self.child_list.append(data)
        else:
            raise TypeError("Can only add int/double/basecclass")


    def __len__(self):
        return self._size()


    def free(self):
        self._free( )



import datetime
from ecl.summary import EclSum


def mock_func(ecl_sum , key , days):
    return days * 10


def createEclSum( case,
                  keys,
                  sim_start = datetime.date(2010 , 1, 1),
                  data_start = None,
                  sim_length_days = 5 * 365,
                  num_report_step = 5,
                  num_mini_step = 10,
                  dims = (20,10,5) ,
                  func_table = {},
                  restart_case = None,
                  restart_step = -1):

    ecl_sum = EclSum.restart_writer(case , restart_case, restart_step, sim_start , dims[0] , dims[1] , dims[2])
    var_list = []
    for (kw,wgname,num,unit) in keys:
        var_list.append( ecl_sum.addVariable( kw , wgname = wgname , num = num, unit =unit) )

    # This is a bug! This should not be integer division, but tests are written
    # around that assumption.
    report_step_length = float(sim_length_days // num_report_step)
    mini_step_length = float(report_step_length // num_mini_step)

    if data_start is None:
        time_offset = 0
    else:
        dt = data_start - sim_start
        time_offset = dt.total_seconds() / 86400.0

    for report_step in range(num_report_step):
        for mini_step in range(num_mini_step):
            days = time_offset + report_step * report_step_length + mini_step * mini_step_length
            t_step = ecl_sum.addTStep( report_step + 1 , sim_days = days )

            for var in var_list:
                key = var.getKey1( )
                if key in func_table:
                    func = func_table[key]
                    t_step[key] = func( days )
                else:
                    t_step[key] = mock_func( ecl_sum , key , days)

    return ecl_sum



#  Copyright (C) 2017  Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from __future__ import print_function
import warnings
from ecl.summary.ecl_sum_node import EclSumNode


class EclSumVector(object):
    def __init__(self, parent, key, report_only = False):
        """
        A summary vector with a vector of values and time.

        A summary vector contains the the full time history of one
        key, along with the corresponding time vectors in several
        different time formats. Depending on the report_only argument
        the data vectors in the EclSumVector can either contain all
        the time values, or only those corresponding to report_steps.

        The EclSumVector contains a reference to the parent EclSum
        structure and this is used to implement several of the
        properties and methods of the object; the EclSum vector
        instances should therefore only be instantiated through the
        EclSum.get_vector() method, and not manually with the
        EclSumVector() constructor.
        """
        self.parent = parent
        self.key = key
        self.report_only = report_only

        if report_only:
            warnings.warn("The report_only flag to the EclSumVector will be removed", DeprecationWarning)

        self.__dates = parent.get_dates(report_only)
        self.__days = parent.get_days(report_only)
        self.__mpl_dates = parent.get_mpl_dates(report_only)
        self.__report_step = parent.get_report_step(report_only)
        self.__values = None


    def __str__(self):
        return "<Summary vector: %s>" % self.key

    def __repr__(self):
        return 'EclSumVector(key = %s, size = %d, unit = %s)' % (self.key, len(self), self.unit)

    @property
    def unit(self):
        """
        The unit of this vector.
        """
        return self.parent.unit(self.key)

    def assert_values(self):
        """
        This function will load and internalize all the values.
        """
        if self.__values is None:
            self.__values = self.parent.numpy_vector(self.key, report_only = self.report_only)

    @property
    def values(self):
        """
        All the summary values of the vector, as a numpy vector.
        """
        self.assert_values()
        return self.__values

    @property
    def dates(self):
        """
        All the dates of the vector, list of datetime() instances.
        """
        return self.__dates

    @property
    def days(self):
        """
        The time in days as a numpy vector.

        In the case of lab unit this will be hours.
        """
        return self.__days

    @property
    def mpl_dates(self):
        """
        All the dates as numpy vector of dates in matplotlib format.
        """
        return self.__mpl_dates

    @property
    def report_step(self):
        """
        All the report_step of the vector.
        """
        return self.__report_step


    def __iget(self, index):
        """
        Will return an EclSumNode for element @index; should be called
        through the [] operator, otherwise you can come across
        unitialized data.
        """
        return EclSumNode(self.__report_step[index],
                          self.__days[index],
                          self.__dates[index],
                          self.__mpl_dates[index],
                          self.__values[index])


    def __len__(self):
        """
        The length of the vector - used for the len() builtin.
        """
        return len(self.__days)


    def __getitem__(self, index):
        """
        Implements the [] operator.

        Will return EclSumNode instance according to @index. The index
        value will be interpreted as in a normal python [] lookup,
        i.e. negative values will be interpreted as starting from the
        right and also slice notation is allowed[*].

        [*] Observe that in the case of slices the return value will
            not be a proper EclSumVector instance, but rather a normal
            Python list of EclSumNode instances.
        """
        self.assert_values()
        length = len(self.values)
        if isinstance(index, int):
            if index < 0:
                index += len(self.__values)
            if index < 0 or index > length:
                raise KeyError("Invalid index:%d out of range [0:%d)" % (index, length))
            else:
                return self.__iget(index)
        elif isinstance(index, slice):
            # Observe that the slice based lookup does __not__ return
            # a proper EclSumVector instance; it will merely return
            # a simple Python list with EclSumNode instances.
            (start, stop, step) = index.indices(length)
            index = start
            sub_vector = []
            while index < stop:
                sub_vector.append(self.__iget(index))
                index += step
            return sub_vector

        raise KeyError("Invalid index:%s - must have integer or slice." % index)

    @property
    def first(self):
        """
        Will return the first EclSumNode in this vector.
        """
        self.assert_values()
        return self.__iget(0)

    @property
    def last(self):
        """
        Will return the last EclSumNode in this vector.
        """
        self.assert_values()

        index = len(self.__values) - 1
        return self.__iget(index)

    @property
    def last_value(self):
        """
        Will return the last value in this vector.
        """
        self.assert_values()

        index = len(self.__values) - 1
        return self.__iget(index).value


    def get_interp(self, days=None, date=None):
        """
        Will lookup value interpolated to @days or @date.

        The function requires one, and only one, time indicator in
        terms of @days or @date. If the @date variable is given that
        should be Python datetime instance.

           vec = sum["WWCT:A-3"]
           vec.get_interp(days = 100)
           vec.get_interp(date = datetime.date(year, month, day))

        This function will crash and burn if the time arguments are
        invalid; if in doubt you should check first.
        """
        return self.parent.get_interp(self.key, days, date)


    def get_interp_vector(self, days_list=None, date_list=None):
        """
        Will return Python list of interpolated values.

        See get_interp() for further details.
        """
        return self.parent.get_interp_vector(self.key, days_list, date_list)


    def get_from_report(self, report_step):
        """
        Will lookup the value based on @report_step.
        """
        return self.parent.get_from_report(self.key, report_step)

    #################################################################

    def first_gt_index(self, limit):
        """
        Locates first index where the value is above @limit.

        Observe that this method will raise an exception if it is
        called from a vector instance with report_only = True.
        """
        if not self.report_only:
            key_index = self.parent._get_general_var_index(self.key)
            time_index = self.parent._get_first_gt(key_index, limit)
            return time_index
        else:
            raise Exception("Sorry - first_gt_index() can not be called for vectors with report_only=True")

    def first_gt(self, limit):
        """
        Locate the first EclSumNode where value is above @limit.

           vec = sum["WWCT:A-3"]
           w = vec.first_gt(0.50)
           print('Water cut above 0.50 in well A-3 at: %s' % w.date)

        Uses first_gt_index() internally and can not be called for
        vectors with report_only = True.
        """
        time_index = self.first_gt_index(limit)
        print(time_index)
        if time_index >= 0:
            return self.__iget(time_index)
        else:
            return None

    def first_lt_index(self, limit):
        """
        Locates first index where the value is below @limit.

        See first_gt_index() for further details.
        """
        if not self.report_only:
            key_index = self.parent._get_general_var_index(self.key)
            time_index = self.parent._get_first_lt(key_index, limit)
            return time_index
        else:
            raise Exception("Sorry - first_lt_index() can not be called for vectors with report_only=True")

    def first_lt(self, limit):
        """
        Locates first element where the value is below @limit.

        See first_gt() for further details.
        """
        time_index = self.first_lt_index(limit)
        if time_index >= 0:
            return self.__iget(time_index)
        else:
            return None



#  Copyright (C) 2017  Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from __future__ import absolute_import, division, print_function, unicode_literals
from six import string_types
from cwrap import BaseCClass
from ecl.util.util import monkey_the_camel
from ecl.util.util import CTime
from ecl import EclPrototype

class EclFileView(BaseCClass):
    TYPE_NAME             = "ecl_file_view"
    _iget_kw              = EclPrototype("ecl_kw_ref    ecl_file_view_iget_kw( ecl_file_view , int)")
    _iget_named_kw        = EclPrototype("ecl_kw_ref    ecl_file_view_iget_named_kw( ecl_file_view , char* , int)")
    _get_unique_kw        = EclPrototype("char*         ecl_file_view_iget_distinct_kw( ecl_file_view, int )")
    _get_size             = EclPrototype("int           ecl_file_view_get_size( ecl_file_view )")
    _get_num_named_kw     = EclPrototype("int           ecl_file_view_get_num_named_kw( ecl_file_view , char* )")
    _get_unique_size      = EclPrototype("int           ecl_file_view_get_num_distinct_kw( ecl_file_view )")
    _create_block_view    = EclPrototype("ecl_file_view_ref ecl_file_view_add_blockview( ecl_file_view , char*, int )")
    _create_block_view2   = EclPrototype("ecl_file_view_ref ecl_file_view_add_blockview2( ecl_file_view , char*, char*, int )")
    _restart_view         = EclPrototype("ecl_file_view_ref ecl_file_view_add_restart_view( ecl_file_view , int, int, time_t, double )")


    def __init__(self):
        raise NotImplementedError("Can not instantiate directly")


    def __iget(self, index):
        return self._iget_kw(index).setParent(parent=self)


    def __repr__(self):
        return 'EclFileView(size=%d) %s' % (len(self), self._ad_str())

    def iget_named_kw(self, kw_name, index):
        if not kw_name in self:
            raise KeyError("No such keyword: %s" % kw_name)

        if index >= self.numKeywords(kw_name):
            raise IndexError("Too large index: %d" % index)

        return self._iget_named_kw(kw_name, index).setParent(parent=self)



    def __getitem__(self, index):
        """
        Implements [] operator; index can be integer or key.

        Will look up EclKW instances from the current EclFile
        instance. The @index argument can either be an integer, in
        which case the method will return EclKW number @index, or
        alternatively a keyword string, in which case the method will
        return a list of EclKW instances with that keyword:

           restart_file = ecl_file.EclFile("ECLIPSE.UNRST")
           kw9 = restart_file[9]
           swat_list = restart_file["SWAT"]

        The keyword based lookup can be combined with an extra [] to
        get EclKW instance nr:

           swat9 = restart_file["SWAT"][9]

        Will return the 10'th SWAT keyword from the restart file. The
        following example will iterate over all the SWAT keywords in a
        restart file:

           restart_file = ecl_file.EclFile("ECLIPSE.UNRST")
           for swat in restart_file["SWAT"]:
               ....
        """

        if isinstance(index, int):
            ls = len(self)
            idx = index
            if idx < 0:
                idx += ls
            if 0 <= idx < ls:
                return self.__iget(idx)
            else:
                raise IndexError('Index must be in [0, %d), was: %d.' % (ls, index))

        if isinstance(index, slice):
            indices = index.indices(len(self))
            kw_list = []
            for i in range(*indices):
                kw_list.append(self[i])
            return kw_list
        else:
            if isinstance(index, bytes):
                index = index.decode('ascii')
            if isinstance(index, string_types):
                if index in self:
                    kw_index = index
                    kw_list = []
                    for index in range(self.numKeywords(kw_index)):
                        kw_list.append( self.iget_named_kw(kw_index, index))
                    return kw_list
                else:
                    raise KeyError("Unrecognized keyword:\'%s\'" % index)
            else:
                raise TypeError("Index must be integer or string (keyword)")


    def __len__(self):
        return self._get_size()


    def __contains__(self, kw):
        if self.numKeywords(kw) > 0:
            return True
        else:
            return False


    def num_keywords(self, kw):
        return self._get_num_named_kw(kw)


    def unique_size(self):
        return self._get_unique_size()


    def unique_kw(self):
        return [self._get_unique_kw(index) for index in range(self.unique_size())]


    def block_view2(self, start_kw, stop_kw, start_index):
        idx = start_index
        if start_kw:
            if not start_kw in self:
                raise KeyError("The keyword:%s is not in file" % start_kw)

            ls = self.numKeywords(start_kw)
            if idx < 0:
                idx += ls
            if not (0 <= idx < ls):
                raise IndexError('Index must be in [0, %d), was: %d.' % (ls, start_index))

        if stop_kw:
            if not stop_kw in self:
                raise KeyError("The keyword:%s is not in file" % stop_kw)

        view = self._create_block_view2(start_kw, stop_kw, idx)
        view.setParent(parent=self)
        return view


    def block_view(self, kw, kw_index):
        num = self.numKeywords(kw)

        if num == 0:
            raise KeyError("Unknown keyword: %s" % kw)

        idx = kw_index
        if idx < 0:
            idx += num

        if not (0 <= idx < num):
            raise IndexError('Index must be in [0, %d), was: %d.' % (num, kw_index))

        view = self._create_block_view(kw, kw_index)
        view.setParent(parent=self)
        return view


    def restart_view(self, seqnum_index=None, report_step=None, sim_time=None, sim_days=None):
        if report_step is None:
            report_step = -1

        if sim_time is None:
            sim_time = -1

        if sim_days is None:
            sim_days = -1

        if seqnum_index is None:
            seqnum_index = -1

        view = self._restart_view(seqnum_index, report_step, CTime(sim_time), sim_days)
        if view is None:
            raise ValueError("No such restart block could be identiefied")

        view.setParent(parent=self)
        return view


monkey_the_camel(EclFileView, 'numKeywords', EclFileView.num_keywords)
monkey_the_camel(EclFileView, 'uniqueSize', EclFileView.unique_size)
monkey_the_camel(EclFileView, 'blockView2', EclFileView.block_view2)
monkey_the_camel(EclFileView, 'blockView', EclFileView.block_view)
monkey_the_camel(EclFileView, 'restartView', EclFileView.restart_view)



#  Copyright (C) 2011  Statoil ASA, Norway. 
#   
#  The file 'hash.py' is part of ERT - Ensemble based Reservoir Tool. 
#   
#  ERT is free software: you can redistribute it and/or modify 
#  it under the terms of the GNU General Public License as published by 
#  the Free Software Foundation, either version 3 of the License, or 
#  (at your option) any later version. 
#   
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY 
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or 
#  FITNESS FOR A PARTICULAR PURPOSE.   
#   
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html> 
#  for more details.
from ctypes import c_void_p

from cwrap import BaseCClass
from ecl import EclPrototype
from ecl.util.util import StringList


class Hash(BaseCClass):
    _alloc =      EclPrototype("void* hash_alloc()" , bind = False)
    _free =       EclPrototype("void hash_free(hash)")
    _size =       EclPrototype("int hash_get_size(hash)")
    _keys =       EclPrototype("stringlist_obj hash_alloc_stringlist(hash)")
    _has_key =    EclPrototype("bool hash_has_key(hash, char*)")
    _get =        EclPrototype("void* hash_get(hash, char*)")
    _insert_ref = EclPrototype("void hash_insert_ref(hash, char*, void*)")

    """
    Base hash class that supports string:void* values
    """

    def __init__(self):
        c_ptr = self._alloc()
        super(Hash, self).__init__(c_ptr)

    def __len__(self):
        return self._size()

    def __getitem__(self, key):
        if self._has_key(key):
            return self._get(key)
        else:
            raise KeyError("Hash does not have key: %s" % key)

    def __setitem__(self, key, value):
        if isinstance(value, c_void_p):
            self._insert_ref(key, value)
        else:
            raise ValueError("Hash does not support type: %s" % value.__class__)

    def __contains__(self, key):
        """ @rtype: bool """
        return self._has_key(key)

    def __iter__(self):
        for key in self.keys():
            yield key

    def keys(self):
        """ @rtype: StringList """
        return self._keys()

    def free(self):
        self._free()

    def __str__(self):
        return str(["%s: %s" % (key, self[key]) for key in self.keys()])


class StringHash(Hash):
    _get_string = EclPrototype("char* hash_get_string(hash, char*)")
    _insert_string = EclPrototype("void hash_insert_string(hash, char*, char*)")

    def __init__(self):
        super(StringHash, self).__init__()

    def __setitem__(self, key, value):
        if isinstance(value, str):
            self._insert_string(key, value)
        else:
            raise ValueError("StringHash does not support type: %s" % value.__class__)

    def __getitem__(self, key):
        if key in self:
            return self._get_string(key)
        else:
            raise KeyError("Hash does not have key: %s" % key)


class IntegerHash(Hash):
    _get_int = EclPrototype("int hash_get_int(hash, char*)")
    _insert_int = EclPrototype("void hash_insert_int(hash, char*, int)")

    def __init__(self):
        super(IntegerHash, self).__init__()

    def __setitem__(self, key, value):
        if isinstance(value, int):
            self._insert_int(key, value)
        else:
            raise ValueError("IntegerHash does not support type: %s" % value.__class__)

    def __getitem__(self, key):
        if key in self:
            return self._get_int(key)
        else:
            raise KeyError("Hash does not have key: %s" % key)


class DoubleHash(Hash):
    _get_double = EclPrototype("double hash_get_double(hash, char*)")
    _insert_double = EclPrototype("void hash_insert_double(hash, char*, double)")

    def __init__(self):
        super(DoubleHash, self).__init__()

    def __setitem__(self, key, value):
        if isinstance(value, int):
            value = float(value)

        if isinstance(value, float):
            self._insert_double(key, value)
        else:
            raise ValueError("DoubleHash does not support type: %s" % value.__class__)

    def __getitem__(self, key):
        if key in self:
            return self._get_double( key)
        else:
            raise KeyError("Hash does not have key: %s" % key)



#  Copyright (C) 2015  Statoil ASA, Norway.
#
#  The file 'ecl_init_file.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from ecl import EclFileEnum
from ecl.eclfile import EclFile, Ecl3DFile


class EclInitFile(Ecl3DFile):

    def __init__(self, grid, filename, flags=0):
        file_type, report_step, fmt_file = EclFile.getFileType(filename)
        if file_type == EclFileEnum.ECL_INIT_FILE:
            super(EclInitFile, self).__init__(grid, filename, flags)
        else:
            err = 'The input filename "%s" does not correspond to a restart file.'
            err += '  Please follow the Eclipse naming conventions.'
            raise ValueError(err % filename)



#  Copyright (C) 2014  Statoil ASA, Norway.
#
#  The file 'fault_line.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from __future__ import print_function
import sys

from ecl.util.util import monkey_the_camel
from ecl.util.util import DoubleVector
from ecl.util.geometry import CPolyline

from .fault_segments import FaultSegment


class FaultLine(object):
    def __init__(self, grid, k):
        self.__grid = grid
        self.__k = k
        self.__segment_list = []
        self.__polyline = None
        self.__ijpolyline = None
        self.__neighborCells = None

    def __len__(self):
        return len(self.__segment_list)


    def __getitem__(self, index):
        return self.__segment_list[index]

    def __iter__(self):
        return iter(self.__segment_list)


    def verify(self):
        if len(self.__segment_list) > 1:
            current = self.__segment_list[0]
            for next_segment in self.__segment_list[1:]:
                if not current.getC2() == next_segment.getC1():
                    sys.stdout.write("Current:   %d ---- %d \n" % (current.getC1()     , current.getC2()))
                    sys.stdout.write("Next   :   %d ---- %d \n" % (next_segment.getC1(), next_segment.getC2()))
                    assert current.getC2() == next_segment.getC1()
                current = next_segment


    def try_append(self, segment):
        if len(self.__segment_list) > 0:
            tail = self.__segment_list[-1]
            if tail.getC2() != segment.getC1():
                if len(self.__segment_list) == 1:
                    if tail.getC2() == segment.getC2():
                        segment.swap()
                    else:
                        tail.swap()
                        if tail.getC2() == segment.getC2():
                            segment.swap()
                else:
                    segment.swap()

            if not tail.getC2() == segment.getC1():
                return False

        self.__segment_list.append(segment)
        self.__polyline = None
        self.__ijpolyline = None
        return True


    def get_k(self):
        return self.__k

    @property
    def k(self):
        return self.__k


    def __init_ij_polyline(self):
        pl = []
        nx = self.__grid.getNX()
        ny = self.__grid.getNY()
        for segment in self:
            corner = segment.getC1()
            i = corner % (nx + 1)
            j = corner // (nx + 1)
            pl.append((i,j))

        segment = self[-1]
        corner = segment.getC2()
        i = corner % (nx + 1)
        j = corner // (nx + 1)
        pl.append((i,j))

        self.__ijpolyline = pl



    def __init_polyline(self):
        pl = CPolyline()
        for (i,j) in self.getIJPolyline():
            x,y,z = self.__grid.getNodeXYZ(i, j, self.__k)
            pl.addPoint(x, y)
        self.__polyline = pl



    def get_polyline(self):
        if self.__polyline is None:
            self.__init_polyline()
        return self.__polyline


    def get_ij_polyline(self):
        if self.__ijpolyline is None:
            self.__init_ij_polyline()
        return self.__ijpolyline



    def __init_neighbor_cells(self):
        self.__neighborCells = []
        nx = self.__grid.getNX()
        ny = self.__grid.getNY()
        k  = self.__k

        for segment in self:
            (j1,i1) = divmod(segment.getC1(), (nx + 1))
            (j2,i2) = divmod(segment.getC2(), (nx + 1))

            if j1 > j2:
                j1,j2 = j2,j1

            if i1 > i2:
                i1,i2 = i2,i1


            if i1 == i2:
                i = i1
                for j in range(j1, j2):
                    g2 = i + j * nx + k * nx*ny
                    if i == 0:
                        g1 = -1
                    else:
                        g1 = g2 - 1

                    if i == nx:
                        g2 = -1

                    self.__neighborCells.append((g1,g2))
            elif j1 == j2:
                j = j1
                for i in range(i1,i2):
                    g2 = i + j * nx + k * nx*ny
                    if j == 0:
                        g1 = -1
                    else:
                        g1 = g2 - nx

                    if j == ny:
                        g2 = -1

                    self.__neighborCells.append((g1,g2))
            else:
                raise Exception("Internal error: found fault segment with "
                                "variation in two directions")

            self.__neighborCells.sort(key=lambda k: (k[0], k[1]))


    def get_neighbor_cells(self):
        if self.__neighborCells is None:
            self.__init_neighbor_cells()

        return self.__neighborCells


    def center(self):
        xlist = DoubleVector()
        ylist = DoubleVector()
        for segment in self:
            C1 = segment.getC1()
            C2 = segment.getC2()
            (J1, I1) = divmod(C1, self.__grid.getNX() + 1)
            (J2, I2) = divmod(C2, self.__grid.getNX() + 1)

            (x1,y1,z) = self.__grid.getNodePos(I1, J1, self.__k)
            (x2,y2,z) = self.__grid.getNodePos(I2, J2, self.__k)

            xlist.append(x1)
            xlist.append(x2)

            ylist.append(y1)
            ylist.append(y2)


        N = len(xlist)
        return (xlist.elementSum()/N, ylist.elementSum()/N)



    def reverse(self):
        reverse_list = reversed(self.__segment_list)
        self.__segment_list = []
        for segment in reverse_list:
            C1 = segment.getC1()
            C2 = segment.getC2()

            rseg = FaultSegment(C2, C1)
            self.tryAppend(rseg)


    def start_point(self):
        pl = self.getPolyline()
        return pl[0]

    def end_point(self):
        pl = self.getPolyline()
        return pl[-1]

    def dump(self):
        print('-----------------------------------------------------------------')
        for segment in self:
            C1 = segment.getC1()
            C2 = segment.getC2()
            (J1, I1) = divmod(C1, self.__grid.getNX() + 1)
            (J2, I2) = divmod(C2, self.__grid.getNX() + 1)
            print('[Corner:%5d IJ:(%3d,%d)] -> [Corner:%5d IJ:(%3d,%d)]'
                  % (C1, I1, J1, C2, I2, J2))


monkey_the_camel(FaultLine, 'tryAppend', FaultLine.try_append)
monkey_the_camel(FaultLine, 'getK', FaultLine.get_k)
monkey_the_camel(FaultLine, 'getPolyline', FaultLine.get_polyline)
monkey_the_camel(FaultLine, 'getIJPolyline', FaultLine.get_ij_polyline)
monkey_the_camel(FaultLine, 'getNeighborCells', FaultLine.get_neighbor_cells)
monkey_the_camel(FaultLine, 'startPoint', FaultLine.start_point)
monkey_the_camel(FaultLine, 'endPoint', FaultLine.end_point)
